/**
 * Module containing basic capabilities of an elementary nature.
 * @author Andre Platzer
 * @version 2003-08-01
 */
:- module basecap.
:- interface.
:- import_module string,
	char, list, set, map, stack, assoc_list, std_util.

/**
 * Cast objects of type T to objects of type S.
 * Fail if conversion is not possible.
 */
:- func cast(T) = S.
:- mode cast(in) = out is semidet.

/**
 * Cast objects of type T to objects of type S.
 * Errors if conversion is not possible.
 */
:- func cast_det(T, string) = S.
:- mode cast_det(in, in) = out is det.


/**
 * Typeclass for types supporting the formation of a a string representation
 * via to_string(T)=string
 * @author Andre Platzer
 * @version 2003-08-01
 */
:- typeclass printable(T) where [
        func to_string(T) = string,
	mode to_string(in) = out is det
].

:- instance printable(int).
:- instance printable(char).
:- instance printable(string).
:- instance printable(univ).

:- instance printable(list(T)) <= printable(T).
:- instance printable(set(T)) <= printable(T).
:- instance printable(map(K,V)) <= (printable(K),printable(V)).
:- instance printable(stack(T)) <= printable(T).

:- instance printable(pair(A,B)) <= (printable(A),printable(B)).
:- instance printable({A,B}) <= (printable(A),printable(B)).


/**
 * semidet_to_det(f) is a function like f but expected to be deterministic,
 * resulting in an error in case of failure.
 * semidet_to_det(f,e) behaves the same, but outputs a specific error e in case of
 * determinism failure.
 */
:- func semidet_to_det(func(A)=B) = (func(A)=B) <= printable(A).
:- mode semidet_to_det(in(func(in)=out is semidet)) = out(func(in)=out is det) is det.
:- func semidet_to_det(func(A)=B, string) = (func(A)=B) <= printable(A).
:- mode semidet_to_det(in(func(in)=out is semidet), in) = out(func(in)=out is det) is det.

:- func semidet_to_det_unprintable(func(A)=B, string) = (func(A)=B).
:- mode semidet_to_det_unprintable(in(func(in)=out is semidet), in) = out(func(in)=out is det) is det.

/**
 * det_to_semidet(f) weakens the determinism specifier.
 */
:- func det_to_semidet(func(A)=B) = (func(A)=B).
:- mode det_to_semidet(in(func(in)=out is det)) = out(func(in)=out is semidet) is det.

/**
 * Get some solution of a nondeterministic predicate.
 * Which one is in principle completely undefined. However, implementations
 * usually guarantee coming up with the first solution produced by the predicate.
 * @see std_util#do_while/2
 */
:- func some_solution(pred(T)) = T.
:- mode some_solution(in(pred(out) is nondet)) = out is semidet.
/*
  Software Error: sorry, not implemented: taking address of pred
 `first_sols/2' with multiple modes.
 (use an explicit lambda expression instead)
:- mode some_solution(in(pred(out) is semidet)) = out is semidet.
:- mode some_solution(in(pred(out) is det)) = out is det.
:- mode some_solution(in(pred(out) is multi)) = out is det.
*/

/**
 * Find all indices where X occurs in L.
 */
:- pred index(T::in,list(T)::in, int::out) is nondet.

/**
 * assoc_list_to_func(AL) converts an associative list to a semidet function.
 */
:- func assoc_list_to_func(assoc_list(K,V)) = (func(K)=V).
:- mode assoc_list_to_func(in) = out(func(in)=out is semidet) is det.

/**
 * map_to_func(AL) converts a map to a semidet function.
 */
:- func map_to_func(map(K,V)) = (func(K)=V).
:- mode map_to_func(in) = out(func(in)=out is semidet) is det.

:- func map_to_func2(map({K1,K2},V)) = (func(K1,K2)=V).
:- mode map_to_func2(in) = out(func(in,in)=out is semidet) is det.

:- func det_map_from_assoc_list(assoc_list(K,V)) = map(K,V)
	<= (printable(K),printable(V)).
:- func det_map_from_assoc_list_unprintable(assoc_list(K,V)) = map(K,V).


/**
 * Casts an arbitrary type into a type that has been promised to be printable.
 * Undefined behaviour if T is not printable.
 */
:- some [Printable] (func unsafe_promise_printable(T) = Printable
		    => printable(Printable)).

/**
 * State-less impure printing for debug.
 * Behaves like io__write_string without being embedded into io__state transitions.
 */
:- /*impure*/ pred log_print(string).
:- mode log_print(in) is det.



:- implementation.
:- import_module require, std_util, functionals.
:- use_module bool.


cast(X) = Y :-
	univ_to_type(univ(X), Y).
cast_det(X,Msg) = Y :-
	cast(X) = Y
	;
	error(Msg).

semidet_to_det(F,S) = (func(X) = Y :-
		(
		  F(X) = Y
		;
		  report_lookup_error("function expected to be deterministic failed and is only semideterministic: "
				     ++ S ++ "\nfunc:" ++ type_name(type_of(F)) ++ " on\nKey: " ++ to_string(X) ++ " of type " ++ type_name(type_of(X)),
				      X, Y)
		)
		).
semidet_to_det(F) = semidet_to_det(F,type_name(type_of(F))).
det_to_semidet(F) = (func(X::in) = (F(X)::out) is semidet).

semidet_to_det_unprintable(F,S) = (func(X) = Y :-
		(
		  F(X) = Y
		;
		  report_lookup_error("function expected to be deterministic failed and is only semideterministic: "
				     ++ S,
				      X, Y)
		)
		).

:- pred first_sols(pred(T),list(T)).
:- mode first_sols(in(pred(out) is nondet),out) is cc_nondet.
/*:- mode first_sols(in(pred(out) is det),out) is det.
:- mode first_sols(in(pred(out) is semidet),out) is semidet.
:- mode first_sols(in(pred(out) is multi),out) is det.*/
first_sols(P,L) :-
	do_while(P,
		 pred(X::in,bool__no::out,_::in,[X]::out) is det,
		 [],
		 L
		).
some_solution(P) = R :-
	%%[R | _] = solutions(P).
	[R] = promise_only_solution(first_sols(P)).



:- instance printable(int) where [
	func(to_string/1) is int_to_string
].
:- instance printable(char) where [
	func(to_string/1) is char_to_string
].
:- instance printable(string) where [
	func(to_string/1) is id
].

:- instance printable(univ) where [
	to_string(U) = "printing universals is not implemented"
	%%(to_string(U) = to_string(V) :- univ_to_type(U,V))
].

:- instance printable(list(T)) <= printable(T) where [
	to_string(L) = "[" ++ join_list(", ", map(to_string, L)) ++ "]"
 ].

:- instance printable(set(T)) <= printable(T) where [
	to_string(C) = "{" ++ join_list(", ", to_sorted_list(map(to_string, C))) ++ "}"
 ].

:- instance printable(map(K,V)) <= (printable(K),printable(V)) where [
	to_string(M) = "<" ++ foldl((func(X,Y,S) = R :-
				    Pair = to_string(X) ++ ": " ++ to_string(Y) ++ "\n",
				    (S = "" ->
					R = Pair
				    ;
					R = S ++ "  " ++ Pair
				    )
				    ), M, "") ++ ">"
 ].

:- instance printable(stack(T)) <= printable(T) where [
	(to_string(S) = R :-
	is_empty(S) ->
	    R=""
	;
	    (
	      pop(S,X,S2),
	      R = to_string(X) ++ ", " ++ to_string(S2))
	    ;
	      error("not is_empty so pop cannot fail")
	    )
].

:- instance printable(pair(A,B)) <= (printable(A),printable(B)) where [
	to_string((X - Y)) = to_string(X) ++ "-" ++ to_string(Y)
 ].

:- instance printable({A,B}) <= (printable(A),printable(B)) where [
	to_string({X,Y}) = "(" ++ to_string(X) ++ "-" ++ to_string(Y) ++ ")"
 ].


assoc_list_to_func(AL) = (func(K::in) = (V::out) is semidet :- search(AL,K,V)).

map_to_func(M) = search(M).
map_to_func2(Map) = (func(X::in,Y::in) = (map__search(Map, {X,Y})::out) is semidet).

%% @internal could be optimized
index(X,L, R) :-
	member(R, 1 `..` length(L)),
	index1_det(L, R, X).

basecap__det_map_from_assoc_list_unprintable(L) = det_insert_from_assoc_list(map__init, L).

basecap__det_map_from_assoc_list([]) = map__init.
basecap__det_map_from_assoc_list([(K-V)|List]) = Map2 :-
	Map = basecap__det_map_from_assoc_list(List),
	(map__contains(Map, K) ->
	    error("duplicate key " ++ to_string(K)
		 ++ " with values " ++ to_string(map__lookup(Map,K))
		 ++ " and " ++ to_string(V))
	;
	    Map2 = map__det_insert(Map, K, V)
	).


:- pragma promise_pure(log_print/1).
:- pragma c_code(log_print(S::in),
        [may_call_mercury],
        "puts(S);").

:- pragma promise_pure(unsafe_promise_printable/1).
:- pragma c_code(unsafe_promise_printable(X::in) = (P::out),
        [may_call_mercury],
        "P = X;").
	

:- end_module basecap.