/**
 * Dynamically transformed set.
 * @author Andre Platzer
 * @version 2003-08-19
 * @see set
 */
:- module dynatraf_set.
:- interface.
:- import_module set, list.

:- type dynatraf_set(T).

/**
 * processEach(p,A, S0, S)
 * S := S0
 * repeat
 *   remove any X from A
 *   p(X,NewA, S,S')
 *   if p has not failed then
 *       A := A union NewA
 *       S := S'
 *   fi
 * until A = []
 * return S
 * The last two arguments of P and processEach are for DCG style accumulators.
 * @preconditions A and NewA are usually assumed, but not strictly required, to contain no duplicates.
 * @internal use to implement the foreachinc construct
 *  which applies precisely once for a dynamically extending set
 * @internal introduced a foreachinc data structure that applies the trafo to each added element exactly once, simplifying closure/1 and generate/8.
 */
:- pred processEach(pred(T,list(T), S, S), list(T), S, S).
:- mode processEach(in(pred(in,out, in,out) is semidet), in, in, out) is det.

/**
 * Like processEach(p,A, S0, S), but enforcing that no element is processed twice.
 * Removes any elements from NewA that are already in S.
 */
:- pred processEachUnique(pred(T,list(T), set(T), set(T)), set(T), set(T)).
:- mode processEachUnique(in(pred(in,out, in,out) is semidet), in, out) is det.

/**
 * Create a set that dynamically transforms all its elements precisely once
 * with the given transformation function.
 */
/*:- func init(func(T)=T) = dynatraf_set(T).

:- func to_set(dynatraf_set(T)) = set(T).

:- func insert(dynatraf_set(T), T) = dynatraf_set(T).
*/


:- implementation.

:- import_module require, std_util.

processEach(P, A, In, Out) :-
	processEachHelper(P,
			  A, Remaining,
			  In, Out),
	(
	  Remaining = []
	;
	  error("ToProcess list has been fully processed")
	).

:- pred processEachHelper(pred(T,list(T), S,S),
			  list(T), list(T),
			  S, S).
:- mode processEachHelper(in(pred(in,out, in,out) is semidet),
			  in, out,
			  in, out) is det.
processEachHelper(P,
		  ToProcess, NewToProcess,
		  In, Out) :-
	(   ToProcess = [],
	    NewToProcess = ToProcess,
	    In = Out
	;
	    ToProcess = [X|ToProcessRest],
	    ( P(X,NewA, In, Out1) ->
		Out2 = Out1,
		ToProcess2 = merge_and_remove_dups(ToProcessRest,NewA)
	    ;
		Out2 = In,
		ToProcess2 = ToProcessRest
	    ),
	    processEachHelper(P,
			      ToProcess2,NewToProcess,
			      Out2,Out)
	).


processEachUnique(P, In, Out) :-
	processEach((pred(X::in,NewA::out, S::in,Sp::out) is semidet :-
		     P(X,NewA1, S, Sp),
		     NewA = to_sorted_list(filter(isnt(contains(S)), list_to_set(NewA1)))
		    ),
		    to_sorted_list(In),
		    In, Out).

:- end_module dynatraf_set.