/**
 * @(#)functionals.java 0.9 2000/07/25 Andre Platzer
 *
 * Copyright (c) 2000 Andre Platzer. All Rights Reserved.
 */

:- module functionals.

/**
 * Provides important compositional functionals for functors.
 * For that, this module contains several static methods that work like <i>Functionals</i>,
 * i.e. high-order functions that have a Function in their signature.
 *
 * @version 0.9, 2000/07/25
 * @author  Andre Platzer
 */
:- interface.
:- import_module list.

    %% // atomic recursion schemes

    /**
     * catamorphism(c, f)
     * Folds a list with a BinaryFunction.
     * @param c the right (second) argument to start with.
     *        The result of the application of f will progressively build the next right (second) argument.
     * @param f the function used to fold the list a with.
     * @param a the list of arguments to be iteratively used as the left (first) arguments.
     * @return (|c,f|): A<sup>*</sup>&rarr;B.
     */
    :- func catamorphism(B, (func(A, B) = B)) = (func(list(A)) = B).
    :- mode catamorphism(in, in(func(in, in) = out is det)) = out(func(in) = out is det) is det.

     /**
     * anamorphism(g, p)
     * Anamorphism recursion functional operator (lense).
     *
     * @param g is a function g:B&rarr;A&times;B=A||B.
     * @param p is a Predicate p:B&rarr;Boole.
     * @return [(g,p)]: B&rarr;A<sup>*</sup>
     */
    :- func anamorphism((func(B) = {A, B}), pred(B)) = (func(B) = list(A)).
    :- mode anamorphism(in(func(in) = out is det), in(pred(in) is semidet)) = out(func(in) = out is det) is det.

     /**
     * hylomorphism((c,f), (g,p))
     * Hylomorphism recursion functional operator (envelope).
     *
     * @param c is an element in C.
     * @param f is a function f:B||C&rarr;C.
     * @param g is a function p:A&rarr;B||A.
     * @param p is a predicate p:A&rarr;Boole.
     * @return |[(c,f),(g,p)]|: A&rarr;C
     * @todo convert arguments to pair or tuple {c,f}
     */
    :- func hylomorphism(C, (func(B, C) = C), (func(A) = {B, A}), pred(A)) = (func(A) = C).
    :- mode hylomorphism(in, in(func(in, in) = out is det), in(func(in) = out is det), in(pred(in) is semidet)) = out(func(in) = out is det) is det.
    :- inst tuple(I1, I2) == bound({I1, I2}).
    :- func hylomorphism({C, (func(B, C) = C)}, {(func(A) = {B, A}), pred(A)}) = (func(A) = C).
    :- mode hylomorphism(in(tuple(ground, func(in, in) = out is det)), in(tuple(func(in) = out is det, pred(in) is semidet))) = out(func(in) = out is det) is det.

    /**
     * paramorphism(b, f, a, r)
     * Paramorphism recursion functional operator (barbed wire).
     * @param b the basevalue to use.
     * @param f a binary function f:A&times;(A*||B)&rarr;B.
     * @return {|b, f|}: A<sup>*</sup>&rarr;B
     */
    :- func paramorphism(B, (func(A, list(A), B) = B)) = (func(list(A)) = B).
    :- mode paramorphism(in, in(func(in, in, in) = out is det)) = out(func(in) = out is det) is det.
    :- func paramorphismi(B, (func(int, B) = B)) = (func(int) = B).
    :- mode paramorphismi(in, in(func(in, in) = out is det)) = out(func(in) = out is det) is det.

    /**
     * nest(functor, n)
     * Nests a function n times within itself.
     * @return functor<sup>n</sup>:A&rarr;A; x&#8614;functor(...functor(functor(x))...)
     * @see std_util.pow/3
     */
    :- func nest((func(A) = A), int) = (func(A) = A).
    :- mode nest(in(func(in) = out is det), in) = out(func(in) = out is det) is det.
    %:- mode nest(in(func(in) = out is semidet), in) = out(func(in) = out is semidet) is det.
    %:- mode nest(in(func(in) = out is multi), in) = out(func(in) = out is multi) is det.
    %:- mode nest(in(func(in) = out is nondet), in) = out(func(in) = out is nondet) is det.
    %:- mode nest(in(func(in) = out is failure), in) = out(func(in) = out is failure) is det.


    /**
     * Performs a fixed-point iteration according to Kleene.
     * Starts applying f repeatedly, starting from a until nothing changes any more.
     * (Fixed-point theorem of Kleene)
     * Let T be completely ordered with =<, i.e. T has a minimum and
     * each monotonously ascending chain has a supremum in T, and if
     * f:T->T is (chain-)continuous, then
     * &exists; sup<sub>k</sub> f<sup>k</sup>(min T) = min fix f
     */
    :- func fixedPoint(T, func(T)=T) = T.
    :- mode fixedPoint(in, in(func(in)=out is det)) = out is det.
    %%:- mode fixedPoint(in, in(func(in)=out is semidet)) = out is semidet.

    /**
     * fixed-point for nondeterministic functions, in case of don't care choices of
     * the transition value.
     * Makes any one transition to a different value until that is no longer possible.
     */
    :- func fixedPointMulti(T, func(T)=T) = T.
    :- mode fixedPointMulti(in, in(func(in)=out is nondet)) = out is det.


    /**
     * &iota;x.p denotes the unique object x for which p is true, if that exists.
     * Undefined if no or more than one such x exists.
     * @see basecap#some_solution/1
     */
    :- func iota(pred(T)) = T.
  /*
  Software Error: sorry, not implemented: taking address of pred
 `first_sols/2' with multiple modes.
 (use an explicit lambda expression instead)
    :- mode iota(pred(out) is det) = out is det.
    :- mode iota(pred(out) is semidet) = out is semidet.
    :- mode iota(pred(out) is multi) = out is semidet.
  */
    :- mode iota(pred(out) is nondet) = out is semidet.

    /**
     * &epsilon;x.p denotes some object x for which p is true, if that exists.
     * Undefined if no x exists. When there are multiple x, it is undefined
     * which one &epsilon;x.p denotes. &epsilon; is the Hilbert-&epsilon operator.
     * @see basecap#some_solution/1
     */
    :- func epsilon(pred(T)) = T.
  /*
  Software Error: sorry, not implemented: taking address of pred
 `first_sols/2' with multiple modes.
 (use an explicit lambda expression instead)
    :- mode epsilon(pred(out) is det) = out is det.
    :- mode epsilon(pred(out) is multi) = out is det.
    :- mode epsilon(pred(out) is semidet) = out is semidet.
 */
    :- mode epsilon(pred(out) is nondet) = out is semidet.



:- implementation.
:- import_module require, basecap.
:- import_module int.

    /**
     * foldRight(+F, +C, +A, ?R)
     * Folds a list with a BinaryFunction.
     * @param f the function used to fold the list a with.
     * @param c the right (second) argument to start with.
     *        The result of the application of f will progressively build the next right (second) argument.
     * @param a the list of arguments to be iteratively used as the left (first) arguments.
     * @return (|c,f|) a.
     */
    catamorphism(C, F) = catamorphismHelper(C, F).
    /*catamorphism(C, F) = Cata :-
        Cata = (func(List) = R :-
            /*List=[], R=C
            ; List=[First|Rest], R=F(First, Cata(Rest))*/
            /*
            catamorphismHelper(C, F, List) = R
        ).*/
    
    :- func catamorphismHelper(B, (func(A, B) = B), list(A)) = B.
    :- mode catamorphismHelper(in, in(func(in, in) = out is det), in) = out is det.
    catamorphismHelper(C, _, []) = C.
    catamorphismHelper(C, F, [First|Rest]) = F(First, catamorphismHelper(C, F, Rest)).

    anamorphism(G, P) = Ana :-
        Ana = (func(B) = List :-
            /*( P(B) -> List=[]
            ; {A, B2} = G(B), List=[A,Ana(B2)]
            )*/
            anamorphismHelper(G, P, B, List)
        ).
    :- pred anamorphismHelper((func(B) = {A, B}), pred(B), B, list(A)).
    :- mode anamorphismHelper(in(func(in) = out is det), in(pred(in) is semidet), in, out) is det.
    anamorphismHelper(G, P, B, R) :-
        (P(B) ->
            R = []
        ;
            {A, B1} = G(B),
            anamorphismHelper(G, P, B1, HB1),
            R = [A| HB1]
        ).
    /*anamorphismHelper(_, P, B, []) :-
        P(B).
    anamorphismHelper(G, P, B, [A| HB1]) :-
        not P(B),
        {A, B1} = G(B),
        anamorphismHelper(G, P, B1, HB1).*/


    hylomorphism(C,F, G,P) = Hylo :-
        Hylo = (func(A) = R :-
            /*(P(A) -> R = C
            ; {B, A2} = G(A), R = F(B, Hylo(A2))
            )*/
            hylomorphismHelper(C,F, G,P, A, R)
        ).
    hylomorphism({C,F}, {G,P}) = hylomorphism(C,F, G,P).

    :- pred hylomorphismHelper(C, (func(B, C) = C), (func(A) = {B, A}), pred(A), A, C).
    :- mode hylomorphismHelper(in, in(func(in, in) = out is det), in(func(in) = out is det), in(pred(in) is semidet), in, out) is det.
    hylomorphismHelper(C,F, G,P, A, R) :-
        (P(A) ->
            R = C
        ;
            {B, A1} = G(A),
            hylomorphismHelper(C,F, G,P, A1, HA1),
            R = F(B, HA1)
        ).
    /*hylomorphismHelper(C,_, _,P, A, C) :-
        P(A).
    hylomorphismHelper(C,F, G,P, A, R) :-
        not P(A),
        (B - A1) = G(A),
        hylomorphismHelper(C,F, G,P, A1, HA1),
        F(B,HA1) = R.*/

    paramorphism(B, F) = paramorphismHelper(B, F).
    :- func paramorphismHelper(B, (func(A, list(A), B) = B), list(A)) = B.
    :- mode paramorphismHelper(in, in(func(in, in, in) = out is det), in) = out is det.
    paramorphismHelper(B, _, []) = B.
    paramorphismHelper(B, F, [A|As]) = R :-
        HAs = paramorphismHelper(B, F, As),
        R = F(A, As, HAs).
    paramorphismi(B, F) = paramorphismiHelper(B, F).
    :- func paramorphismiHelper(B, (func(int, B) = B), int) = B.
    :- mode paramorphismiHelper(in, in(func(in, in) = out is det), in) = out is det.
    paramorphismiHelper(B, F, N) = R :-
        require(N >= 0, "non-negative integer expected for paramorphismi"),
        (N = 0 ->
            R = B
        ;
            %%N > 0,
            N1 = N - 1,
            HN = paramorphismiHelper(B, F, N1),
            R = F(N1, HN)
        ).
    /*paramorphismi(B, _, 0, B).
    paramorphismi(B, F, N, R) :-
        N > 0,
        paramorphismi(B, F, N - 1, HN),
        R = F(N, HN).*/

    nest(Functor, N) = FN :-
        require(N >= 0, "non-negative nesting expected"),
        ( N = 0 ->
            FN = (func(X) = X)
        ;   
            %FN = (func(X) = nest(Functor, N - 1)(Functor(X)) )
            FN = (func(X) = Functor( nest(Functor, N - 1)(X) ))
        ).


fixedPoint(A, F) = R :-
	fixedPointHelper(A, F(A), F, R).
:- pred fixedPointHelper(T, T, func(T)=T, T).
:- mode fixedPointHelper(in,in,in(func(in)=out is det),out) is det.
fixedPointHelper(OldA, A, F, R) :-
	(OldA = A ->
	    R = A
	;
	    fixedPointHelper(A, F(A), F, R)
	).


fixedPointMulti(A, F) = R :-
	(FA = some_solution((pred(X::out) is nondet :- X=F(A), X\=A)) ->
	  fixedPointMultiHelper(A, FA, F, R)
	;
	  R = A
	).
:- pred fixedPointMultiHelper(T, T, func(T)=T, T).
:- mode fixedPointMultiHelper(in,in,in(func(in)=out is nondet),out) is det.
fixedPointMultiHelper(OldA, A, F, R) :-
	(OldA = A ->
	    R = A
	;
	    (FA = some_solution((pred(X::out) is nondet :- X=F(A), X\=A)) ->
	      (
		FA \= A
	      ;
		error("solution satisfies defining equation")
	      ),
	      fixedPointMultiHelper(A, FA, F, R)
	    ;
	      R = A
	    )
	).


%% @todo check that the solution is unique. Or at least check when in assertion mode
iota(P) = some_solution(P).

epsilon(P) = some_solution(P).

%@todo import Functionals.java stuff

:- end_module functionals.