/**
 * Provides common grammar analysis functionality.
 * @author Andre Platzer
 * @version 2003-08-10
 */
:- module grammar_analysis.
:- interface.

:- import_module scanners, parsers, list, set, basecap.

%% grammar flow analysis routines

/**
 * Performs a bottom-up grammar flow analysis.
 * bottom_up_grammar_flow_analysis(G,min,F,C)
 * @param G the grammar to analyse.
 * @param min the initial minimum value to start the analysis with.
 * @param F the transfer function used to transfer the information from the
 *  right hand side <em>nonterminals</em> of a production to the left hand side nonterminal.
 *  Every call F(p,v) will satisfy length(v)=length(rhs(p))
 * @param C the combination function used to combine the separate information
 *  from all right hand sides of one particular nonterminal.
 * @return
 *  The solution I:N&rarr;Value of the recursive equation system
 *  I(X) = C({F(X->X0&rarr;u0 X1 u1 X2 u2 ... Xn un, I(X1),...,I(Xn))
 *       &brvbar; X->X0&rarr;u0 X1 u1 X2 u2 ... Xn un &isin; &Pi;}) &forall;X&isin;N
 */
:- func bottom_up_grammar_flow_analysis(grammar, Value,
					func(production,list(Value))=Value,
					func(set(Value))=Value
				       ) = (func(symbol)=Value)
	<= printable(Value).
:- mode bottom_up_grammar_flow_analysis(in, in,
					in(func(in,in)=out is det),
					in(func(in)=out is det)
				       ) = out(func(in)=out is det) is det.

/**
 * Performs a top-down grammar flow analysis.
 * top_down_grammar_flow_analysis(G,IS,min,F,C)
 * @param G the grammar to analyse.
 * @param min the initial minimum value to start the analysis with.
 * @param IS the value for the starting symbol of the grammar.
 * @param F the transfer function used to transfer the information from the
 *  left hand side nonterminal of a production to the occurrence i
 *  of a nonterminal on the right hand side. Here the occurrence is the index
 *  of its occurrence in the right hand side including terminals, starting from 1.
 *  Every call F(p,i,v) will satisfy rhs(p)(i):nonterminal(_)
 * @param C the combination function used to combine the separate information
 *  from all occurrences of one particular nonterminal in all right hand sides.
 * @return
 *  The solution I:N&rarr;Value of the recursive equation system
 *  I(S) = IS
 *  I(X) = C({F(Y->p1...pn,i, I(Y)) with pi=X and Y->p1...pn &isin; &Pi;}) &forall;X&isin;N\{S}
 */
:- func top_down_grammar_flow_analysis(grammar, Value, Value,
				       func(production,int,Value)=Value,
				       func(set(Value))=Value
				      ) = (func(symbol)=Value)
	<= printable(Value).
:- mode top_down_grammar_flow_analysis(in, in, in,
				       in(func(in,in,in)=out is det),
				       in(func(in)=out is det)
				      ) = out(func(in)=out is det) is det.

%% applications

/**
 * FIRSTk determines for each nonterminal X the k prefixes of any terminal words
 * derived from it, i.e.
 * FIRSTk(&alpha;) = k:FIRST(&alpha;)
 * FIRST(&alpha;) = {u &brvbar; &alpha; &rArr;<sup>*</sup> u}
 * @return the FIRSTk function on symbols.
 * @note Reduces tokens of terminals to their token_kind.
 */
:- func first(int,grammar) = (func(symbol)=set(word_kind)).
:- mode first(in,in) = out(func(in)=out is det) is det.

/**
 * FIRSTk determines the k prefixes of any terminal words derived from alpha, i.e.
 * @note Reduces tokens of terminals to their token_kind.
 */
:- func first(int,grammar,symbol_chain) = set(word_kind).
:- mode first(in,in,in) = out is det.


/**
 * FOLLOWk determines for each nonterminal X the k prefixes of any terminal words
 * that can follow X, i.e.
 * FOLLOWk(X) = FIRSTk(FOLLOW(X))
 * FOLLOW(X) = {&beta; &brvbar; &exist;&alpha; S &rArr;<sup>*</sup> &alpha;X&beta;}
 * @note Reduces tokens of terminals to their token_kind.
 */
:- func follow(int,grammar) = (func(symbol)=set(word_kind)).
:- mode follow(in,in) = out(func(in)=out is det) is det.


/**
 * prefix(k,L) computes the k-prefixes of the words formed by concatenation
 * of L(1)L(2)L(3)...L(n).
 */
:- func prefix(int,list(set(word(Token)))) = set(word(Token)).



:- implementation.

:- import_module scanners, basecap, map, assoc_list, string, std_util, functionals, require, int.

%% tools

/**
 * Complete information about nonterminals to information about all symbols for a production.
 * complete_terminals(C,p,Vs)
 * Given a way C to complete the value for individual terminals,
 * supplements the value information for the nonterminals Vs to
 * an integrated value list holding at each position C(a) for terminals a
 * and Vs(X) for nonterminals X according to the production.
 * @param C the function to call for completion of terminal information.
 * @param p the production under examination, or more precisely, only its right hand side.
 * @param Vs the information for nonterminals.
 * @return integrated information about terminals and nonterminals according to C and Vs.
 */
:- func complete_for_terminals(symbol_chain,(func(token_kind)=Value),list(Value)) = list(Value)
	<= printable(Value).
:- mode complete_for_terminals(in,in(func(in)=out is det),in) = out is det.



:- func bottom_up_grammar_flow_analysisHelper(grammar, list(pair(symbol,Value)), func(production,list(Value))=Value, func(set(Value))=Value) = assoc_list(symbol,Value)
	<= printable(Value).
:- mode bottom_up_grammar_flow_analysisHelper(in, in, in(func(in,in)=out is det), in(func(in)=out is det)) = out(func(in)=out is det) is det.
bottom_up_grammar_flow_analysisHelper(G, I0, F, C) = Istar :-
	P = list_to_set(productions(G)),
        Istar = fixedPoint(I0,
	  (func(I) = map((func(X) =
			 pair(X,
			    C(filter_map(
				  (func(rule(Y,Alpha,COMP)::in)
				  = (R::out)
				  is semidet
				  :- R
				  = F(rule(Y,Alpha,COMP),
				      map(semidet_to_det(assoc_list_to_func(I),"current state lookup in " ++ to_string(I)),
					  filter(is_nonterminal, Alpha)
					 )
				     ),
				  X = Y/*,
					  log_print(" combining " ++ to_string(rule(Y,Alpha,COMP))
						   ++ " to "++ to_string(R)
						   ++ "\n   over " ++ to_string(map(semidet_to_det(assoc_list_to_func(I),"current state lookup in " ++ to_string(I)),
										    Alpha))
						   ++ "  from " ++ to_string(Alpha)
						   ++ "\n   in " ++ to_string(I))
					 */
				  ),
				  P
				 )
			     )
			   )
			 ),
			 keys(I))
	  %%:- log_print("fixed-point iteration " ++ to_string(I) ++ " for " ++ type_name(type_of(F)))
	  )
			  ).

:- func bottom_up_grammar_flow_analysis_map(grammar, Value,
					func(production,list(Value))=Value,
					func(set(Value))=Value
				       ) = map(symbol,Value)
	<= printable(Value).
:- mode bottom_up_grammar_flow_analysis_map(in, in,
					in(func(in,in)=out is det),
					in(func(in)=out is det)
				       ) = out is det.
bottom_up_grammar_flow_analysis_map(G, V0, F, C)
 = map__optimize(map__from_assoc_list(
  bottom_up_grammar_flow_analysisHelper(G,
				  map((func(X)=pair(X,V0)), set__to_sorted_list(nonterminals(G))),
				  F,
				  C)
		      )).

bottom_up_grammar_flow_analysis(G, V0, F, C)
= semidet_to_det(map_to_func(
			     bottom_up_grammar_flow_analysis_map(G, V0, F, C)
			    ),
		 "combine fixed-point map to function").



:- func top_down_grammar_flow_analysisHelper(grammar, Value, Value,
				       func(production,int,Value)=Value,
				       func(set(Value))=Value
				      ) = assoc_list(symbol,Value)
	<= printable(Value).
:- mode top_down_grammar_flow_analysisHelper(in, in, in,
				       in(func(in,in,in)=out is det),
				       in(func(in)=out is det)
				      ) = out is det.
top_down_grammar_flow_analysisHelper(G, V0, IS, F, C) = Istar :-
	P = list_to_set(productions(G)),
	I0 = map((func(X)=pair(X,V) :-
		     X = start(G) ->
		      V = IS
		     ;
		      V = V0),
		     set__to_sorted_list(nonterminals(G))),
        Istar = fixedPoint(I0,
	  (func(I) = map((func(X) = X-V
			 :-
			 X = start(G) ->
			  V = IS
			 ;
			  V = C(solutions_set((pred(R::out) is nondet
					      :-
					       member(rule(Y,Alpha,COMP), P),
					       index(X,Alpha,Index),
					       search(I,Y,IY),
					       R = F(rule(Y,Alpha,COMP),
						     Index,
						     IY
						    )
					      )
					     )
			       )
			 ),
			 keys(I)
			)
	  %%:- log_print("fixed-point iteration " ++ to_string(I) ++ " for " ++ type_name(type_of(F)))
	  )
			  ).

top_down_grammar_flow_analysis(G, V0, IS, F, C)
= semidet_to_det(map_to_func(map__optimize(map__from_assoc_list(
				    top_down_grammar_flow_analysisHelper(G,
				       V0,
				       IS,
				       F,
				       C)))),
		 "combine top down fixed-point map to function").


%% tools

:- pred complete_for_terminalsHelper((func(token_kind)=Value),symbol_chain,list(Value),list(Value)).
:- mode complete_for_terminalsHelper(in(func(in)=out is det),in,in,out) is semidet.
complete_for_terminals(Alpha,C,Vs) = R :-
	complete_for_terminalsHelper(C,Alpha,Vs,R)
	;
	error("terminal completion impossible because of a mismatch between rule and values\n"
	     ++ "complete for " ++ to_string(Alpha) ++ " on values " ++ to_string(Vs)).

complete_for_terminalsHelper(_,[],[],[]).
complete_for_terminalsHelper(C,[Ai|Alpha],Vs,[R|Rs]) :-
	(
	  Ai = terminal(A),
	  R = C(get_kind(A)),
	  complete_for_terminalsHelper(C,Alpha,Vs,Rs)
	;
	  Ai = nonterminal(_),
	  [R|Vs2] = Vs,
	  complete_for_terminalsHelper(C,Alpha,Vs2,Rs)
	).
	

%% applications

prefix(K,ListOfWords) = R :-
	(
	  ListOfWords = [],
	  R = make_singleton_set([]) %%set__init
	;
	  ListOfWords = [Words|MoreWords],
	  R = power_union(map((
			       func(W) = WP :-
			      WPrefix = take_upto(K,W),
			       (
				 length(WPrefix) < K ->
				 WP = map(append(W), prefix(K-length(WPrefix),MoreWords))
			       ;
				 WP = make_singleton_set(WPrefix)
			       )
			      ), Words)
			 )
	)
	%%,log_print("prefix" ++ to_string(K) ++ " of " ++ to_string(ListOfWords) ++  " = " ++ to_string(R))
	.


/**
 * (set(word),subset,emptyset,union) as demi-lattice
 * F(X->u0 X1 u1 X2...Xn un,d1,...,dn) = k:(u0 d1 u1 d2 ... dn un)
 * C = union
 */
first(K,G) = (func(X) = R :-
	     (
	       X = nonterminal(_),
	       (
		 R = search(FIk, X)
	       ;
		 error("FIRST" ++ to_string(K) ++ " lookup for " ++ to_string(X) ++ " failed")
	       )
	     ;
	       X = terminal(A),
	       (A = end_of_file ->
		   error("what do we do with eof in FIRSTk? {}?")
	       ;
		   R = make_singleton_set([get_kind(A)])
	       )
	     )
	     )
:-
	FIk = bottom_up_grammar_flow_analysis_map(
				   G,
				   set__init,
				   (func(rule(X,Alpha,COMP),FIRSTS)=
				   prefix(K,
					  complete_for_terminals(
								 Alpha,
								 (func(A) = make_singleton_set([A])),
								 FIRSTS
								)
					 )
				   )
				  ,
				   set__power_union
				  ).


first(K,G,ALPHA) = prefix(K,map(first(K,G), ALPHA)).

/**
 * (set(word),subset,emptyset,union) as demi-lattice
 * F(X->u0 X1 u1 X2...Xn un,i,d) = k:(ui FIRSTk(Xi+1) ui+1 FIRSTk(Xi+2)... un d)
 * C = union
 * I(S) = {#}
 */
follow(K,G)
 = top_down_grammar_flow_analysis(
				  G,
				  set__init,
				  make_singleton_set([end_of_file]),
				  (func(rule(X,Alpha,COMP),I,FOLLOWS) = R
				  :-
				  R =
				  prefix(K,
					 map(FIRSTk,
					     Alpha_after_i
					     )
					++ [FOLLOWS]
					)
				  ,
				  %% I+1 because we start after the nonterminal, -1 because indices start with 1, here
				      drop(I,Alpha,Alpha_after_i)
				      /*log_print("  follow " ++ to_string(rule(X,Alpha,COMP)) ++ " @ " ++ to_string(I) ++ " follows " ++ to_string(FOLLOWS) ++ " =>"
					       ++ "\n  rest " ++ to_string(Alpha_after_i) ++ " " ++ to_string([FOLLOWS]) ++ "\n  starting with " ++ to_string(map(FIRSTk,Alpha_after_i) ++ [FOLLOWS])
					       ++ "\n  is " ++ to_string(R))*/
				  ;
				      error("dropping " ++ to_string(I) ++ " from " ++ to_string(Alpha) ++ " should not fail for valid calls by top_down_grammar_analysis")
				  )
				 ,
				  set__power_union
				 )
 :-
	 FIRSTk = first(K,G).

:- end_module grammar_analysis.