/**
 * Grammar notation parsing, i.e. parsing BNF-like notations of grammars.
 * More precisely, in the future this module will parse
 * attributed (relational) definite clause grammars.
 * @author Andre Platzer
 * @version 2003-08-19
 */
:- module grammar_parser.
:- interface.
:- import_module parsers, scanners, scanner_generator.

	/**
	  * parse_grammar(GrammarNotation, Grammar)
	  * Parse the notation of a grammar to an internal grammar representation.
	  */
	:- pred parse_grammar(string, grammar).
	:- mode parse_grammar(in, out) is semidet.

	/**
	  * parse_grammar(GrammarNotation, Grammar, LexicalGrammar)
	  * Parse the notation of a grammar to an internal grammar representation.
	  */
	:- some [Token] (pred parse_grammar(string, grammar, scanner_generator__grammar(Token))
			=> token(Token)).
	:- mode parse_grammar(in, out, out) is semidet.


:- implementation.
:- import_module string, list, require, std_util, basecap, scanners,
	parsers__lr1Parser,
	parser_generator, parser_generator__lr1Generator, parser_generator__sllGenerator,
	scanner_generator,
	string_source,
	simple_tokenizer, token_impl.
:- import_module syntaxTree.
:- use_module scanner_util.

/**
 * Get the parser for the grammar notation grammar.
 */
:- (func grammar_parser = lr1Parser(int)).
:- pragma memo(grammar_parser/0).
grammar_parser = P :-
	TerminalAction = (func([_,ID,_]::in) = (new_action_value(new_terminal(
			/* @todo which token kind to use here? */
			delimiter(action_value_to_type(ID))))::out) is semidet),
	%% the grammar notation grammar along with actions that build up
	%% an internal representation of the grammar parsed
	GG = grammar(nonterminal("Grammar"),
	[
	 new_rule(nonterminal("Grammar"), [nonterminal("Rule"), nonterminal("Grammar")],
		  (func([Prod,Gram]::in)
		    = (new_action_value([action_value_to_type(Prod) `with_type` production
					|action_value_to_type(Gram) `with_type` list(production)
					])::out) is semidet)
	     ),
	 new_rule(nonterminal("Grammar"), [nonterminal("Rule")],
		  (func([Prod]::in)
		    = (new_action_value([action_value_to_type(Prod) `with_type` production])::out) is semidet)
	     ),
	 new_rule(nonterminal("Rule"), [nonterminal("Symbol"), new_terminal(operator("-")), new_terminal(operator("-")), new_terminal(operator(">")), nonterminal("Symbols"), new_terminal(operator("."))],
		  (func([X,_,_,_,Alpha,_]::in)
		    = (new_action_value(rule(action_value_to_type(X),
					     action_value_to_type(Alpha)
					    ))::out) is semidet)
		 ),
	 new_rule(nonterminal("Symbols"), [nonterminal("Symbol"), nonterminal("Symbols")],
		 (func([Y,Alpha]::in)
		  = (new_action_value(
				      [action_value_to_type(Y) `with_type` symbol
				      |action_value_to_type(Alpha) `with_type` list(symbol)
				      ])::out) is semidet)
		 ),
	 new_rule(nonterminal("Symbols"), [],
		 (func([]::in)
		  = (new_action_value([] `with_type` list(symbol))::out) is semidet)
		 ),

	 %% nonterminal case
	 new_rule(nonterminal("Symbol"), [new_terminal(identifier)],
		  (func([ID]::in) = (new_action_value(nonterminal(action_value_to_type(ID)))::out) is semidet)
		 ),
	 %% terminal case
	 new_rule(nonterminal("Symbol"), [new_terminal(delimiter("[")), new_terminal(identifier), new_terminal(delimiter("]"))],
		  (func([_,ID,_]::in) = (new_action_value(new_terminal(
			/* @todo which token kind to use here? */
			make_token_kind(action_value_to_type(ID))))::out) is semidet)
	      ),

	 %% lexical token use case
	 new_rule(nonterminal("Symbol"), [nonterminal("Terminal")],
		  (func([ID]::in) = (ID::out) is semidet)
		 ),

	 %% lexical token definition rule
	 new_rule(nonterminal("Rule"), [nonterminal("Terminal"), new_terminal(delimiter(":")), new_terminal(operator("-")), new_terminal(operator("-")), new_terminal(operator(">")), new_terminal(string_literal), new_terminal(operator("."))],
		  (func([X, _, _,_,_, Alpha,_]::in)
		    = (R::out) is semidet :-
		      %% combine lexical and CH-2 rules into grammar for later splitting.
		      R = new_action_value(rule(action_value_to_type(X),
						[new_terminal(regexp_literal(action_value_to_type(Alpha)))]
					    ))
		  )
		 ),

	 new_rule(nonterminal("Terminal"), [new_terminal(operator("<")), new_terminal(identifier), new_terminal(operator(">"))],
		  (func([_,ID,_]::in) = (new_action_value(new_terminal(make_token_kind(action_value_to_type(ID))))::out) is semidet)
		 ),

	 %% lexical token definition and use rule
	 new_rule(nonterminal("Symbol"), [new_terminal(operator("<")), new_terminal(string_literal), new_terminal(operator(">"))],
		  (func([_,Image,_]::in) = (new_action_value(new_terminal(
			/*no kind please*/make_token_kind(action_value_to_type(Image))))::out) is semidet :-
		  log_print("warning combined token definition and use does not yet work"))
		 ),

	 
	 %% @xxx additional terminal cases needed until token specification is available 
	 new_rule(nonterminal("Symbol"), [new_terminal(delimiter("[")), new_terminal(delimiter("%")), new_terminal(delimiter("]"))],
		  TerminalAction
	      ),
	 new_rule(nonterminal("Symbol"), [new_terminal(delimiter("[")), new_terminal(delimiter("|")), new_terminal(delimiter("]"))],
		  TerminalAction
	      ),
	 new_rule(nonterminal("Symbol"), [new_terminal(delimiter("[")), new_terminal(delimiter("(")), new_terminal(delimiter("]"))],
		  TerminalAction
	      ),
	 new_rule(nonterminal("Symbol"), [new_terminal(delimiter("[")), new_terminal(delimiter(")")), new_terminal(delimiter("]"))],
		  TerminalAction
	      ),
	 new_rule(nonterminal("Symbol"), [new_terminal(delimiter("[")), new_terminal(delimiter("\"")), new_terminal(delimiter("]"))],
		  TerminalAction
	      ),
	 new_rule(nonterminal("Symbol"), [new_terminal(delimiter("[")), new_terminal(delimiter(",")), new_terminal(delimiter("]"))],
		  TerminalAction
	      ),
	 new_rule(nonterminal("Symbol"), [new_terminal(delimiter("[")), new_terminal(delimiter("$")), new_terminal(delimiter("]"))],
		  TerminalAction
	      ),
	 new_rule(nonterminal("Symbol"), [new_terminal(delimiter("[")), new_terminal(delimiter(":")), new_terminal(delimiter("]"))],
		  TerminalAction
	      ),
	 new_rule(nonterminal("Symbol"), [new_terminal(delimiter("[")), new_terminal(delimiter("&")), new_terminal(delimiter("]"))],
		  TerminalAction
	      ),
	 new_rule(nonterminal("Symbol"), [new_terminal(delimiter("[")), new_terminal(operator("+")), new_terminal(delimiter("]"))],
		  TerminalAction
	      ),
	 new_rule(nonterminal("Symbol"), [new_terminal(delimiter("[")), new_terminal(operator("*")), new_terminal(delimiter("]"))],
		  TerminalAction
	      )
	]
	       ),
	(
	  P = lr1_parser_specific(GG)
	;
	  error("internal error: grammar notation grammar")
	).

parse_grammar(GNotation, G) :-
	parse_grammar(GNotation, G, scanner_generator__grammar([])).

parse_grammar(GNotation, G, LexicalGrammar) :-
	TS = scan(simple_tokenizer,string_tokenSequence(GNotation)),
	log_print("Scanned\n" ++ to_string(TS)),
	Output = some_solution(parse(grammar_parser, TS)),
	log_print("Parsed\n" ++ to_string(Output)),
	CombinedProductions = action_value_to_type(Output),
	%% split lexical rules again
	filter((pred(rule(nonterminal(_),_,_)::in) is semidet), CombinedProductions,
	       GrammaticalProductions, IncludedLexicalProductions),
	LexicalProductions = map(semidet_to_det((func(rule(terminal(Tok),[terminal(RegexpTok)],_)::in) =
						(scanner_util__rule_group(get_kind(Tok),to_char_list(Regexp))::out) is semidet :-
						string_literal(Regexp) = cast(RegexpTok)
						),
						"lexical rules included in grammar should have been parsed with a special form"
						),
				 IncludedLexicalProductions),
	LexicalGrammar = grammar(LexicalProductions),
	G = grammar(nt(det_head(GrammaticalProductions)), GrammaticalProductions),
	log_print("lexical grammar " ++ to_string(LexicalProductions)),
	log_print("grammar " ++ to_string(G)).

/**
 * Intermediate solution for making a token of a given regular expressions
 */
:- func regexp_literal(string) = token_impl.
regexp_literal(S) = string_literal(S).

:- end_module grammar_parser.