/**
 * Provides LR(0) parser generator.
 * @author Andre Platzer
 * @version 2003-08-08
 */
:- module parser_generator__lr0Generator.
:- interface.

:- import_module parsers, parsers__lr1Parser.

/**
 * Generate an LR(0) parser.
 * Calculates the parser parameters for the LR parser from an (appropriate) grammar.
 * @param grammar the grammar to parse.
 * @return initialState the initial state of the LR automaton to start parsing.
 * @return actionTable the action table of the LR(0) parser determining the parser action
 *  under some lookahead at a particular state.
 * @return gotoTable the goto table of the LR(0) parser determining the transition under
 *  some symbol in a particular state.
 * @see parsers__lr1Parser#lr1_parse/4
 */
:- some [State] pred lr0_generate(grammar/*(actionType)*/, State, actionTable(State), gotoTable(State))
	=> printable(State).
:- mode lr0_generate(in,out,out(func(in, in) = out is det),out(func(in, in) = out is det)) is semidet.


:- implementation.

:- import_module scanners, basecap, list, set, map, string, std_util, functionals, require.

/**
 * A context-free parser item.
 * item(A,alpha,beta) is depicted [A->alpha.beta]
 * and represents the parser state of having advanced to the position marked
 * by the point while parsing the production
 * A->alphabeta 
 */
:- type item
	---> item(
		  nt :: symbol,
		  history :: list(symbol),
		  future :: list(symbol)
		 ).
:- instance printable(item) where [
	to_string(item(A,Alpha,Beta)) =
	    "[" ++ to_string(A) ++ "->" ++ to_string(Alpha) ++ "." ++ to_string(Beta) ++ "]"
 ].
:- type state==set(item).

/**
 * Create LR(0) PDA = deterministic characteristic finite automaton + PDA
 * // S' is the new starting symbol of the grammar, S the old
 * // thereby assuring that the starting symbol never occurs on the right hand side, and only once
 * func start : state
 *     return {[S'->.S]}
 * 
 * // epsilon closure of s
 * func closure(s:state) : state
 * begin
 *     q := s;
 *     while exist [X->alpha.Ybeta] in q with Y->gamma in Prod and [Y->.gamma] not in q do
 *         q := q union {[Y->.gamma]}
 *     od
 *     return q
 * end
 * 
 * // shift transitions of item PDA
 * func successors(s:state,Y:V):state
 * begin
 *     return {[X->alphaY.beta] with [X->alpha.Ybeta] in s}
 * end
 * 
 * var
 * q,q':state
 * Q:set(state)
 * d:state x V -> state
 * begin
 *     Q := {closure(start)};
 *     d := {};
 *     foreach q in Q and X in V do
 *     // also precisely once for the dynamically extending set Q
 *         q' := closure(successors(q,X));
 * 	   if q' \= {} then
 * 	       if q' not in Q then
 * 	           Q := Q union {q'}
 * 	       fi
 * 	       d := d union {transition q->q' under X}
 * 	   fi
 *     od
 * end
 */

:- func start(grammar) = state.
start(G) = make_singleton_set(item(nonterminal("S'"), [],[start(G)])).

/**
 * // epsilon closure of s
 */
:- func closure(grammar,state) = state.
:- mode closure(in,in) = out is det.
closure(G,S) = R :-
	P = productions(G),
	R = fixedPoint(S,
		       (func(Q) = union(Q,
					power_union(
						   filter_map(
								   func(item(nonterminal(X),Alpha,[nonterminal(Y),Beta]))
								  = filter_map(
										    func(rule(nonterminal(Y),Gamma)) = item(nonterminal(Y),[],Gamma),
										    list_to_set(P)),
								   Q
								  )
						  )
				       )
		       )
		      ).

/**
 * // shift transitions of item PDA
 */
:- func successors(state,symbol) = state.
:- mode successors(in,in) = out is det.
successors(S,Y) =
filter_map((func(item(nonterminal(X),Alpha,[nonterminal(Y)|Beta]))
	   = item(nonterminal(X),Alpha ++ [nonterminal(Y)], Beta)),
	   S
	  ).

/**
 * q,q':state
 * Q:set(state)
 * d:state x V -> state
 * begin
 *     Q := {closure(start)};
 *     d := {};
 *     foreach q in Q and X in V do
 *     // also precisely once for the dynamically extending set Q
 *         q' := closure(successors(q,X));
 * 	   if q' \= {} then
 * 	       if q' not in Q then
 * 	           Q := Q union {q'}
 * 	       fi
 * 	       d := d union {transition q->q' under X}
 * 	   fi
 *     od
 * end
 * @param G the grammar to generate a parser for.
 * @param V the vocabulary of nonterminal and terminal symbols of the grammar G.
 * @param States the current set of states.
 * @return NewStates the new set of states.
 * @param Transitions the current set of transitions pair({q,X},q') from q to q' under X.
 * @return NewTransitions the new set of transitions.
 * @param ToProcess the subset of States which still has to be processed via the foreach loop.
 * @param NewToProcess the remaining subset of states which still has to be processed.
 */
:- pred generate(grammar,set(symbol),
		 set(state),set(pair({state,symbol},state)),set(state),
		 set(state),set(pair({state,symbol},state)),set(state)).
:- mode generate(in,in,
		 in,in,in,
		 out,out,out) is det.
generate(G, V,
	 States, Transitions, ToProcess,
	 NewStates, NewTransitions, NewToProcess) :-
	(set__empty(ToProcess) ->
	    NewStates=States,
	    NewTransitions=Transitions,
	    NewToProcess=ToProcess
	;
	    %% should be [S|ToProcessRest] = ToProcess
	    set__member(S,ToProcess),
	    delete(ToProcess,S,ToProcessRest),
	    
	    News = map(
			      (func(X) = pair(Sp,pair({S,X},Sp)) :-
			      Sp = closure(G,successors(S,X)),
				  not set__empty(Sp)
			      ),
			      V),
	    FoundStates = map(fst, News),
	    States2 = union(States, FoundStates),
	    Transitions2 = union(Transitions, set__map(snd, News)),
	    %% only add those states to the process list that have been new to States.
	    ToProcess2 = union(ToProcessRest,
			       filter((pred(Sp::in) is semidet :- not set__member(Sp,States)),
					   FoundStates)
			      ),
	    generate(G,V,
		     States2,Transitions2,ToProcess2,
		     NewStates,NewTransitions,NewToProcess)
	).

lr0_generate(G, InitialState, ActionTable, GotoTable) :-
	Q0 = closure(G, start(G)),
	generate(G,list_to_set(condense(map(
					  (func(rule(X,Alpha,_)) = [X|Alpha]),
					  productions(G)
					 ))),
		 make_singleton_set(Q0),init,make_singleton_set(Q0),
		 States,Transitions,init),
	InitialState = Q0,
	%% convert Transitions:set({{state,symbol},state}) to proper map
	GotoTableMap = map__from_assoc_list(to_sorted_list(Transitions)),
	GotoTable = ((func(State,Symbol) = R
	       :- R = map__search(GotoTableMap, {State,Symbol})
	       ; error("did not find q=" ++ to_string(State) ++ " a=" ++ to_string(Symbol) ++ " in gotoTable")
	       )
	      `with_type` (func(state,token)=state)),
	ActionTable=[].

:- end_module parser_generator__lr0Generator.