/**
 * Provides LR(1) parser generator.
 * @author Andre Platzer
 * @version 2003-08-09
 */
:- module parser_generator__lr1Generator.
:- interface.

:- import_module parsers, parsers__lr1Parser, basecap.

/**
 * Shift-reduce and reduce-reduce conflict exceptions of an LR(k) parser.
 */
:- type reduce_conflict(Item)
	--->
		/**
		 * Shift-reduce conflict exception of an LR(k) parser.
		 * shift_reduce_conflict(token, action1,item1, action2,item2, furtherDescription)
		 */
		shift_reduce_conflict(token_kind,parserAction,Item,parserAction,Item, string)
	;
		/**
		 * Reduce-reduce conflict exception of an LR(k) parser.
		 * reduce_reduce_conflict(token, action1,item1, action2,item2, furtherDescription)
		 */
		reduce_reduce_conflict(token_kind,parserAction,Item,parserAction,Item, string)
	;
		various_conflict(token_kind,parserAction,parserAction, string).
:- instance printable(reduce_conflict(Item)) <= printable(Item).

/**
 * Generate a canonical LR(1) parser.
 * Calculates the parser parameters for the LR parser from an (appropriate) grammar.
 * @param grammar the grammar to parse.
 * @return initialState the initial state of the LR automaton to start parsing.
 * @return actionTable the action table of the LR(1) parser determining the parser action
 *  under some lookahead at a particular state.
 * @return gotoTable the goto table of the LR(1) parser determining the transition under
 *  some symbol in a particular state.
 * @see parsers__lr1Parser#lr1_parser/4
 */
:- some [State] pred lr1_generate(grammar/*(actionType)*/, State, actionTable(State), gotoTable(State))
	=> printable(State).
:- mode lr1_generate(in,out,out(func(in, in) = out is det),out(func(in, in) = out is det)) is semidet.

/**
 * Create an LR(1) parser.
 * @param grammar the grammar to parse.
 * @see lr1_generate/4
 * @see parsers__lr1Parser#lr1_parser/4
 */
:- some [P] (func lr1_parser(grammar/*(actionType)*/) = P => parser(P)).
:- mode lr1_parser(in) = out is semidet.


%% specific version of lr1_parser for pragma memo to work
:- (func lr1_parser_specific(grammar/*(actionType)*/) = lr1Parser(int)).
:- mode lr1_parser_specific(in) = out is semidet.



:- implementation.

:- import_module scanners, grammar_analysis,
	list, set, map, string,
	basecap, std_util, functionals, require,
	int, dynatraf_set.

%% map versions of actionTable and gotoTable
:- type actionTableMap(State) == map({State,token_kind},parserAction).
:- type gotoTableMap(State) == map({State,symbol}, State).

/**
 * An LR(1) parser item.
 * item(A,alpha,beta,L) is depicted [A->alpha.beta,L]
 * and represents the parser state of having advanced to the position marked
 * by the point while parsing the production
 * A->alphabeta
 * when knowing that L contains the possible tokens of the lookahead
 * (generalizable to list(list(token)) for LR(k) k>1).
 * The item is valid for a reliable prefix gammaalpha if for every u in L
 * there is a right derivation
 * S# &rArr;<sub>R</sub><sup>*</sup> gammaXw &rArr; gammaalphabetaw
 * with k:w=u
 */
:- type item
	---> item(
		  nt :: symbol,
		  history :: list(symbol),
		  future :: list(symbol),
		  lookahead :: set(token_kind)
		 ).
:- instance printable(item) where [
	to_string(item(A,Alpha,Beta,L)) =
	    "[" ++ to_string(A) ++ " -> " ++ join_list(" ", map(to_string, Alpha)) ++ "." ++ join_list(" ", map(to_string,Beta))
		++ ", " ++ to_string(L) ++ "]"
 ].
:- type state==set(item).
:- type encoded_state==int.

/**
 * item_lookahead(X,Alpha,Beta, Q, L)
 * sets L to the lookahead of the(!) item(X,Alpha,Beta,L) in Q.
 * Get the lookahead of the specified item in a state, knowing that
 * states only contain one lookahead for each item X->Alpha.Beta.
 */
:- pred item_lookahead(symbol,list(symbol),list(symbol), set(token_kind), state).
:- mode item_lookahead(in,in,in, out, in) is semidet.
item_lookahead(X, Alpha, Beta, L, Q) :-
	L = some_solution(
			  (pred(Lookahead::out) is nondet :-
			  member(item(X, Alpha,Beta, Lookahead), Q)
			  )
			 ).

:- instance printable(reduce_conflict(Item)) <= printable(Item) where [
	to_string(shift_reduce_conflict(A, Act1,Item1, Act2,Item2, Description)) =
		   "grammar is not LR(1)\nconflict " ++ to_string(Act1) ++ "\n   in\t " ++ to_string(Item1) ++ "\n versus  " ++ to_string(Act2) ++ "\n   in\t " ++ to_string(Item2) ++ "\n  under\t " ++ to_string(A) ++ "\n" ++ to_string(Description) ++ "\n",
	to_string(reduce_reduce_conflict(A, Act1,Item1, Act2,Item2, Description)) =
		   "grammar is not LR(1)\nconflict " ++ to_string(Act1) ++ "\n   in\t " ++ to_string(Item1) ++ "\n versus  " ++ to_string(Act2) ++ "\n   in\t " ++ to_string(Item2) ++ "\n  under\t " ++ to_string(A) ++ "\n" ++ to_string(Description) ++ "\n",
	to_string(various_conflict(A, Act1, Act2, Description)) =
		   "grammar is not LR(1)\ndiffuse  " ++ to_string(Act1) ++ "\n versus  " ++ to_string(Act2) ++ "\n  under\t " ++ to_string(A) ++ "\n" ++ to_string(Description) ++ "\nThis should never happen."
 ].

/**
 * Encodes the states of an LR(1) parser.
 * @param initialState the initial state of the LR automaton to start parsing.
 * @param actionTable the action table of the LR(1) parser determining the parser action
 *  under some lookahead at a particular state.
 * @param gotoTable the goto table of the LR(1) parser determining the transition under
 *  some symbol in a particular state.
 * @return encodedInitialState the initial state of the LR automaton to start parsing.
 * @return encodedActionTable the action table of the LR(1) parser determining the parser action
 *  under some lookahead at a particular state.
 * @return encodedGotoTable the goto table of the LR(1) parser determining the transition under
 *  some symbol in a particular state.
 */
:- all [State] /*some [EncodedState]*/ ((
	pred lr1_encode(set(State), State, actionTableMap(State), gotoTableMap(State),
						   encoded_state, actionTableMap(encoded_state), gotoTableMap(encoded_state))
				    %%=> printable(EncodedState)
				    ) <= printable(State)).
:- mode lr1_encode(in, in,in,in, out,out,out) is det.



%% @see basecap#map_to_func2
:- func map_to_actionTable(actionTableMap(State)) = actionTable(State)
	<= printable(State).
:- mode map_to_actionTable(in) = out(func(in,in) = out is det) is det.
map_to_actionTable(ActionTableMap) = ((func(State,Symbol) = R
				  :- R = map__search(ActionTableMap, {State,Symbol})
				     %%,log_print(">>\tlookup acti " ++ to_string({State,Symbol}) ++ " : " ++ to_string(R))%% ++ " in\n\t" ++ to_string(ActionTableMap) ++ "<<")
				  ; error("Syntax error: no action in state " ++ to_string(State) ++ " with lookahead " ++ to_string(Symbol))
				  )
				 `with_type` (func(State,token_kind)=parserAction)).

:- func map_to_gotoTable(gotoTableMap(State)) = gotoTable(State)
	<= printable(State).
:- mode map_to_gotoTable(in) = out(func(in,in) = out is det) is det.
map_to_gotoTable(GotoTableMap) = ((func(State,Symbol) = R
				  :- R = map__search(GotoTableMap, {State,Symbol})
				     %%,log_print(">>\tlookup goto " ++ to_string({State,Symbol}) ++ " : " ++ to_string(R))%% ++ " in\n\t" ++ to_string(GotoTableMap) ++ "<<")
				  ; error("Syntax error: no transition in state " ++ to_string(State) ++ " under " ++ to_string(Symbol))
				  )
				 ).

/**
 * Generate the LR(1) action table from the states.
 *  for each q in Q do
 *      for each K in q do
 *          if K=[S'->S.,{#}] then
 *              action(q,#) := accept
 *          else if K=[X->alpha.,L] then
 *              for each a in L do
 *                  action(q,a) := reduce(X->alpha)
 *              od
 *          else if K=[X->alpha.abeta,L] then
 *              action(q,a) := shift
 *          fi
 *      od
 *  od
 * @throws (currently instead of failing) error in case of shift-reduce or reduce-reduce conflicts.
 */
:- func lr1_generateActionTable(grammar,set(state)) = actionTable(state).
:- mode lr1_generateActionTable(in,in) = out is semidet.
lr1_generateActionTable(G,States) = map_to_actionTable(lr1_generateActionTableMap(G,States)).

:- func lr1_generateActionTableMap(grammar,set(state)) = actionTableMap(state).
:- mode lr1_generateActionTableMap(in,in) = out is semidet.
lr1_generateActionTableMap(G,States) = ActionTableMap :-
	solutions(
		  (pred(pair({Q,A},Act)::out) is nondet :-
		      member(Q,States),
		      member(K,Q),
		      (   K = item(nonterminal("S'"), [start(G)],[], make_singleton_set(end_of_file `with_type` token_kind)),
			  A = end_of_file,
			  Act = accept
		      ;
			  K = item(nonterminal(X), Alpha,[], L),
			  member(A,L),
			  member(rule(nonterminal(X), Alpha, COMP), productions(G)),
			  Act = reduce(rule(nonterminal(X), Alpha, COMP))
		      ;
			  K = item(_, _,[terminal(A0)|_], _),
			  A = get_kind(A0),
			  A \= end_of_file,
			  Act = shift
		      )
		  )
		 ,TableEntries
		  ),
	%% improved error messages, instead of uninformative failing in det_map_from_assoc_list
	(
	  error(to_string(some_solution((pred(R::out) is nondet :-
		   list__member({Q,A}-Act1, TableEntries),
		   list__member({Q,A}-Act2, TableEntries),
		   %% conflicting actions?
		   %% No shift-shift conflicts, please. They are, of course,
		   %% implicitly handled by the deterministic subset construction of the characteristic finite automaton, already, and just an artefact of this test
		   %% No identical reduce actions, please.
		   %% @internal here we need explicit unification to avoid higher-order unification error at run-time
		   not parserAction_equals(Act1, Act2),
		   (
		     Act1 = shift,
		     %% how to handle duplicates resulting from K1 and K2 changing places?
		     %% find items causing conflict
		     member(K1,Q),
		     K1 = item(_, _,[terminal(A0)|_], _),
		     A = get_kind(A0),
		     member(K2,Q),
		     K2 = item(_, _Gamma,[], L),
		     member(A,L),
		     R = shift_reduce_conflict(A, Act1,K1, Act2,K2, " in state " ++ to_string(Q))
		   ;
		     Act1 = reduce(rule(nonterminal(X),Alpha1,_)),
		     %% if Act2 was shift, we already handled with K1 and K2 changing places
		     Act2 \= shift,
		     %% find items causing conflict
		     Act2 = reduce(rule(nonterminal(X2),Alpha2,_)),
		     member(K1,Q),
		     K1 = item(nonterminal(X), Alpha1,[], L1),
		     member(K2,Q),
		     K2 = item(nonterminal(X2), Alpha2,[], L2),
		     not set__empty(intersect(L1,L2)),
		     R = reduce_reduce_conflict(A, Act1,K1, Act2,K2, " in state " ++ to_string(Q))
		   ;
		     Act1 \= shift,
		     Act1 \= reduce(_),
		     R = various_conflict(A, Act1,Act2," in state " ++ to_string(Q))
		   )
		  )
				  )
			 ))
	;
	  ActionTableMap = map__optimize(det_map_from_assoc_list(TableEntries))
	).

/**
 * &epsilon;-free FIRST1
 */
:- func epsilon_ffi(grammar,symbol_chain,set(token_kind)) = set(token_kind).
epsilon_ffi(G, Alpha, L) = R :-
	R = filter_map(
		       (func([A|_RestWord]::in) = (A::out) is semidet),
		       prefix(1,
			      [first(1,G,Alpha),
			       map((func(A)=[A]),L)]
			     )
		      )
	.%%,log_print("   epsilon_ffi(" ++ to_string(Alpha) ++ ", " ++ to_string(L) ++ ") = " ++ to_string(R) ++ "\n   FIRST1(" ++ to_string(Alpha) ++ ")=" ++ to_string(first(1,G,Alpha))).


/**
 * Create LR(1) PDA = deterministic characteristic finite automaton + PDA
 * // S' is the new starting symbol of the grammar, S the old
 * // thereby assuring that the starting symbol never occurs on the right hand side, and only once
 * func start : state
 *     return {[S'->.S,{#}]}
 * 
 * // epsilon closure of s=q
 * func closure(s:state) : state
 * begin
 *     foreachinc [X->alpha.Ybeta,L] in q and Y->gamma in Prod do
 *         if exist [Y->.gamma,L'] in q then
 *             replace [Y->.gamma,L'] by [Y->.gamma,L' union epsilon-ffi(betaL)]
 *         else
 *             q := q union {[Y->.gamma,epsilon-ffi(betaL)]}
 *         fi
 *     od
 *     return q
 * end
 * 
 * // shift transitions of item PDA
 * func successors(s:state,Y:V):state
 * begin
 *     return {[X->alphaY.beta,L] with [X->alpha.Ybeta,L] in s}
 * end
 * 
 * var
 * q,q':state
 * Q:set(state)
 * d:state x V -> state
 * begin
 *     Q := {closure(start)};
 *     d := {};
 *     foreachinc q in Q and Y in V do
 *         q' := closure(successors(q,Y));
 * 	   if q' \= {} then
 * 	       if q' not in Q then
 * 	           Q := Q union {q'}
 * 	       fi
 * 	       d := d union {transition q->q' under Y}
 * 	   fi
 *     od
 * end
 */

:- func start(grammar) = state.
start(G) = make_singleton_set(item(nonterminal("S'"), [],[start(G)], make_singleton_set(end_of_file))).

/**
 * Form the closure under &epsilon;-transitions of s, i.e. add all items
 * that could be reached by expanding items of s at their current positions.
 */
:- func closure(grammar,state) = state.
:- mode closure(in,in) = out is det.
closure(G,S) = Sp :-
	P = productions(G),
	processEachUnique((pred(item(nonterminal(X),Alpha,[nonterminal(Y)|Beta], L)::in,
			  ToProcessTotal::out,
			  in, out) is semidet -->
		    %% for each Y->gamma
		    foldl2((pred(Prod::in, ToProcess1::in,NewToProcess1::out, Q::in,Qp::out) is det :-
			   ( Prod = rule(nonterminal(Y),Gamma,COMP) ->
			       %%log_print(" ... epsclose " ++ to_string(rule(nonterminal(Y),Gamma,COMP)) ++ " from " ++ to_string(item(nonterminal(X),Alpha,[nonterminal(Y)|Beta], L)) ++ " leads to " ++ to_string(epsilon_ffi(G,Beta,L)) ++ "\n\ttodo\t" ++ to_string(ToProcess1)),
			       (item_lookahead(nonterminal(Y), [],Gamma, Lp, Q) ->
				   %% @todo optimize could combine lookup and delete here
				   delete(Q, item(nonterminal(Y), [],Gamma, Lp), Q2),
				   NewK = item(nonterminal(Y), [],Gamma, union(Lp,epsilon_ffi(G,Beta,L))),
				   insert(Q2,NewK, Qp)
			       ;
				   NewK = item(nonterminal(Y), [],Gamma, epsilon_ffi(G,Beta,L)),
				   insert(Q,NewK,Qp)
			       ),
			       NewToProcess1 = [NewK | ToProcess1]
			   ;
			       %% rule does not fit, leave things as they are
			       Qp = Q,
			       NewToProcess1 = ToProcess1
			   )
			   ),
			   P,			   
			   [], ToProcessTotal
			   )
		    ),
		    S,
		    Sp
		   ).

/**
 * // shift transitions of item PDA
 */
:- func successors(state,symbol) = state.
:- mode successors(in,in) = out is det.
successors(S,Y) =
filter_map((func(item(nonterminal(X), Alpha,[Y2|Beta], L)::in)
	   = (item(nonterminal(X), Alpha ++ [Y2], Beta, L)::out) is semidet :-
	   (
	     Y = nonterminal(_),
	     Y = Y2
	   ;
	     Y = terminal(A),
	     Y2= terminal(A2),
	     token_class_equals(A,A2)
	   )
	   %%,log_print("   immed.succ\t" ++ to_string(item(nonterminal(X), Alpha,[Y2|Beta], L)) ++ " under " ++ to_string(Y) ++ " is\n\t\t" ++ to_string(item(nonterminal(X), Alpha ++ [Y2], Beta, L)))
	   ),
	   S
	  ).

/**
 * Generate 
 * @param G the grammar to generate a parser for.
 * @param States the current set of states.
 * @return NewStates the new set of states.
 * @param Transitions the current set of transitions pair({q,X},q') from q to q' under X.
 * @return NewTransitions the new set of transitions.
 * @param ToProcess the subset of States which still has to be processed via the foreach loop.
 * @param NewToProcess the remaining subset of states which still has to be processed.
 */
:- pred generate(grammar,
		 set(state),set(pair({state,symbol},state)),
		 set(state),set(pair({state,symbol},state))).
:- mode generate(in,
		 in,in,
		 out,out) is det.
generate(G,
	 States0, Transitions0,
	 NewStates, NewTransitions) :-
	V = nonterminals(G) `union` map(semidet_to_det((func(terminal(A)::in) = ('new terminal'(get_kind(A))::out) is semidet)), terminals(G)), %%vocabulary(G),
	processEach((pred(S::in,ToProcess::out,
			  {States,Transitions}::in,{States2,Transitions2}::out) is semidet :-
		    News = filter_map(
				      (func(X::in) = (pair(Sp,pair({S,X},Sp))::out) is semidet :-
				          Sp = closure(G,successors(S,X)),
					  not set__empty(Sp)
				      %%,log_print("\nclosure(successors(" ++ to_string(X) ++ "," ++ to_string(S) ++ "))\n\t = closure(" ++ to_string(successors(S,X)) ++ ")\n\t = " ++ to_string(Sp) ++ "\n in states=" ++ to_string(States) ++ "\n")
				      )
				     ,
				      V),
		     FoundStates = map(fst, News),
		     States2 = States `union` FoundStates,
		     Transitions2 = Transitions `union` map(snd, News),
		     %% only add those states to the process list that have been new to States.
		     ToProcess = to_sorted_list(filter(isnt(contains(States)), FoundStates))
		     ),
		    to_sorted_list(States0),
		    {States0,Transitions0}, {NewStates,NewTransitions}).


lr1_generate(G, InitialState, ActionTable, GotoTable) :-
	lr1_generate_specific(G, InitialState, ActionTable, GotoTable).

:- pred lr1_generate_specific(grammar/*(actionType)*/, int, actionTable(int), gotoTable(int)).
:- mode lr1_generate_specific(in,out,out(func(in, in) = out is det),out(func(in, in) = out is det)) is semidet.

lr1_generate_specific(G, InitialState, ActionTable, GotoTable) :-
	Q0 = closure(G, start(G)),
	generate(G,
		 make_singleton_set(Q0), init,
		 States, Transitions),
	log_print("LR(1) " ++ to_string(States)),
	InitialState1 = Q0,
	%% convert Transitions:set({{state,symbol},state}) to proper gotoTableMap
	GotoTableMap1 = map__optimize(det_map_from_assoc_list(to_sorted_list(Transitions))),
	ActionTableMap1 = lr1_generateActionTableMap(G,States),

	log_print("\ngotoTable=\n" ++ to_string(GotoTableMap1) ++ "\nactionTable=\n" ++ to_string(ActionTableMap1)),
	lr1_encode(States, InitialState1, ActionTableMap1, GotoTableMap1,
		   InitialState, ActionTableMap, GotoTableMap),
	GotoTable = map_to_gotoTable(GotoTableMap),
	ActionTable = map_to_actionTable(ActionTableMap),
	log_print("LR(1) gotoTable=\n" ++ to_string(GotoTableMap) ++ "\nactionTable=\n" ++ to_string(ActionTableMap)).

lr1_encode(States,
	   InitialState, ActionTable, GotoTable,
	   EncInitialState, EncActionTable, EncGotoTable) :-
	StatesList = to_sorted_list(States),
	Encode = (func(S) = I-1
		 :-
		 nth_member_search(StatesList,S,I)
		 ;
		     error("State not found")
		 ) `with_type` (func(State)=int),
	EncInitialState = Encode(InitialState),
	EncActionTable = map__optimize(det_map_from_assoc_list(map(
						  semidet_to_det((func(({S,L}-A)::in) = (({Encode(S),L}-A)::out) is semidet),
								  "actionTableMap type"),
						  to_assoc_list(ActionTable)
						 ))),
	EncGotoTable = map__optimize(det_map_from_assoc_list(map(
						semidet_to_det((func(({S,A}-Sp)::in) = (({Encode(S),A}-Encode(Sp))::out) is semidet),
							       "gotoTableMap type"),
						to_assoc_list(GotoTable)
					       ))).


lr1_parser(G) = P :-
	lr1_generate(G,I,AT,GT),
	parsers__lr1Parser__lr1_parser(G,I,AT,GT) = P.

lr1_parser_specific(G) = P :-
	lr1_generate_specific(G,I,AT,GT),
	parsers__lr1Parser__lr1_parser_specific(G,I,AT,GT) = P.

:- end_module parser_generator__lr1Generator.