/**
 * Provides SLL(k) parser generator.
 * Especially SLL(1)=LL(1)
 * @author Andre Platzer
 * @version 2003-08-15
 */
:- module parser_generator__sllGenerator.
:- interface.

:- import_module parsers, parsers__sllParser, basecap.

/**
 * Choice conflict exception of an LL(k) parser.
 */
:- type choice_conflict
	---> choice_conflict(production,production,word_kind).
:- instance printable(choice_conflict).

/**
 * Generate an SLL(k) parser.
 * Calculates the parser parameters for the SLL(k) parser from an (appropriate) grammar.
 * @param grammar the grammar to parse.
 * @return parseTable the parser table of the SLL(k) parser predicting the production
 *  to use for a nonterminal under some lookahead terminals.
 * @see parsers__sllParser#sll_parser/3
 */
:- pred sll_generate(grammar/*(actionType)*/, int, parseTable).
:- mode sll_generate(in,in,out(func(in,in)=out is det)) is semidet.

/**
 * Create an SLL(k) parser.
 * @param grammar the grammar to parse.
 * @see sll_generate/2
 * @see parsers__sllParser#sll_parser/3
 */
:- some [P] (func sll_parser(grammar/*(actionType)*/) = P => parser(P)).
:- mode sll_parser(in) = out is semidet.



:- implementation.

:- import_module scanners, grammar_analysis,
	list, set, map, string,
	basecap, std_util, functionals, require,
	int, exception.

:- instance printable(choice_conflict) where [
	to_string(choice_conflict(Prod1,Prod2,LA)) = 
		   "grammar is not SLL(" ++ to_string(length(LA)) ++ "). Choice conflict involving two expansions\nconflict " ++ to_string(Prod1)
		++ "\n versus  " ++ to_string(Prod2) ++ "\n"
		 ++ "a common prefix is " ++ to_string(LA)
 ].


%% map versions of parseTable
:- type parseTableMap == map({symbol,word_kind}, symbol_chain).

:- func map_to_parseTable(parseTableMap) = parseTable.
:- mode map_to_parseTable(in) = out(func(in,in) = out is det) is det.
map_to_parseTable(ParseTableMap) = ((func(X,Lookahead) = R
				  :- R = map__search(ParseTableMap, {X,Lookahead})
				  ; error("Parse exception: for " ++ to_string(X)
					 ++ ", encountered " ++ to_string(Lookahead)
					 ++ ", expected one of " ++ to_string(solutions(
						(pred(LA::out) is nondet :- member({X,LA}, keys(ParseTableMap)))
										)
								      )
					 )
				  )).

sll_generate(G, K, ParseTable) :-
	FIRSTk = first(K,G),
	FOLLOWk = follow(K,G),
	Director = solutions_set((pred(({X,LA}-Alpha)::out) is nondet :-
				 member(rule(X,Alpha,_), productions(G)),
				 member(LA, prefix(K,[FIRSTk(Alpha),FOLLOWk(X)]))
		  )),
	%% improved error messages, instead of uninformative failing in det_map_from_assoc_list
	(
	  error(to_string(some_solution((pred(R::out) is nondet :-
		   member({X,LA}-Alpha1, Director),
		   member({X,LA}-Alpha2, Director),
		   %% conflicting predictions?
		   Alpha1 \= Alpha2,
		   %% how to handle duplicates resulting from Alpha1 and Alpha2 changing places?
		   %% find productions causing conflict
		   member(rule(X,Alpha1,COMP1),productions(G)),
		   member(rule(X,Alpha2,COMP2),productions(G)),
		   R = choice_conflict(rule(X,Alpha1,COMP1), rule(X,Alpha2,COMP2),LA)
		  ))
	       ))
	;
	  ParseTableMap = map__optimize(det_map_from_assoc_list(to_sorted_list(Director))),
	  ParseTable = map_to_parseTable(ParseTableMap),
	  log_print("SLL(" ++ to_string(K) ++ ") " ++ "\nparseTable=\n" ++ to_string(ParseTableMap))
	).

sll_parser(G) = P :-
	MaxK = 1,
	{K,PT} = some_solution((pred({K,K_PT}::out) is nondet :-
			    member(K, 1 `..` MaxK),
			    %% @todo catch exception and retry for next K
			    sll_generate(G,K,K_PT)
		      )
		     ),
	parsers__sllParser__sll_parser(G,K,PT) = P.

:- end_module parser_generator__sllGenerator.