/**
 * Provides a nondeterministic analytic bottom-up parser implementation.
 * Produces right derivations.
 * @author Andre Platzer
 * @version 2003-08-01
 */
:- module parsers__analyticParser.
:- interface.

/**
 * Create the analytic parser.
 * @note in the presence of &epsilon; productions, (depth-first) search does not
 * necessarily terminate.
 */
:- some [P] (func analytic_parser(grammar/*(actionType)*/) = P => parser(P)).
:- mode analytic_parser(in) = out is det.


:- implementation.

:- import_module list, require, std_util, string, basecap, scanners.

:- type analyticParser/*(actionType)*/
	---> analyticParser(
			    grammar :: grammar/*(actionType)*/
			   ).
:- instance printable(analyticParser) where [
	to_string(analyticParser(G)) = "nondeterministic analytic bottom-up parser[" ++ to_string(G) ++ "]"
      ].

:- type stack_symbol/*(actionType)*/
 	---> stack_symbol(
 			  /**
 			   * The symbol that has been pushed to the stack.
 			   * Comes from a finite set.
 			   */
 			  symbol :: symbol,
 			  /**
 			   * The value that has been associated with this stack
 			   * entry during evaluation.
 			   */
 			  value :: actionType
 			 ).

/**
 * reducingParse(Grammar, Tokenstream, WorkingStack, AST)
 * Predict next production.
 * @internal explicit different some T type for stack symbols. Then it could integrate
 * a modifiable data field for keeping the action evaluation.
 */
:- pred reducingParse(analyticParser/*(actionType)*/::in, TS::in, list(stack_symbol/*(actionType)*/)::in, actionType::out) is nondet
	<= tokenSequence(TS).


:- instance parser(analyticParser/*(actionType)*/) where [
	parse(P, TS, Output) :-
		reducingParse(P, TS, [], Output)
  ].
analytic_parser(Grammar) = analyticParser(Grammar).



%%  @internal implementation does not yet support partially instantiated data structures.
%% :- inst stack_symbol_skeleton == bound( stack_symbol(ground, free) ).
%% :- func stackify(symbol) = stack_symbol.
%% :- mode stackify(in) = (free >> stack_symbol_skeleton) is det.
%% stackify(N) = stack_symbol(N,_).

/**
 * pop(Stack, ExpectedSymbols, ValuesFound, RemainingStack)
 * removes length(ExpectedSymbols) stack_symbols from Stack remaining only RemainingStack.
 * Checks that the symbols of those stack_symbols popped off are the ExpectedSymbols
 * and collects the values associated with these in ValuesFound.
 * In other words, pops the top length(ExpectedSymbols) elements from Stack, and
 * splits them into stack_symbol(Symbol,Value) verifying Symbol is the corresponding
 * element in ExpectedSymbols, and adding Value to ValuesFound.
 */
:- pred popExpect(list(stack_symbol/*(actionType)*/), list(symbol),
		  list(actionType), list(stack_symbol/*(actionType)*/)).
:- mode popExpect(in,in,out,out) is semidet.
popExpect(Stack, [], [], Stack).
popExpect([stack_symbol(A, V) | Stack], [A | ExpectedSymbols], [V | ValuesFound], Stack2) :-
	popExpect(Stack, ExpectedSymbols, ValuesFound, Stack2).

/**
 * accept
 * S\phi# ---> \phi
 */
reducingParse(Parser, Tokens, [stack_symbol(start(grammar(Parser)), A)], A) :-
	empty(Tokens).

/**
 * reduce from stack to nonterminal.
 * R\phi ---> X\phi  whenever  (X--->converse(R)) &isin; &Pi;
 */
reducingParse(Parser, Tokens, Stack, Output) :-
	member(Prod, productions(grammar(Parser))),
	Prod = rule(nonterminal(X), RHS, COMP),
	list__reverse(RHS, ReverseRHS),
	popExpect(Stack, ReverseRHS, Values, Stack2),
	%% pop ReverseRHS from Stack
	%% append(ReverseRHS, Stack2, Stack),
	Value = COMP(list__reverse(Values)),
	reducingParse(Parser, Tokens,
		      [stack_symbol(nonterminal(X),Value) | Stack2],
		      Output).

/**
 * shift terminal from input to stack.
 * \phia ---> a\phi
 */
reducingParse(Parser, Tokens, Stack, Output) :-
	get_token(A, Tokens, TokensR),
	%% require(is_terminal(A)),
	reducingParse(Parser, TokensR,
		      [stack_symbol('new terminal'(A), new_action_value(get_value(A))) | Stack],
		      Output).

/**
 * Error handler.
 */
/*reducingParse(_, Tokens, [Delta | _], _) :-
	get_token(Tok, Tokens, _),
	not Tok = Delta,
	error("Syntactic error. stack: " ++ to_string(Delta) ++ ", found: " ++ to_string(Tok)).
*/
