/**
 * Provides a deterministic LR(1) bottom-up parser implementation.
 * Produces right derivations.
 * @author Andre Platzer
 * @version 2003-08-05
 */
:- module parsers__lr1Parser.
:- interface.

:- type parserAction
	---> shift
	;    reduce(production)
	;    accept
	;    error
	. %%%where equality is parserAction_equals.
:- instance printable(parserAction).

:- pred parserAction_equals(parserAction,parserAction).
:- mode parserAction_equals(in,in) is semidet.

/**
 * An action table that associates with each state q and lookahead terminals a
 * the action that the parser should do.
 * @preconditions q:State & a:list(token)
 *  & mode actionTable(in,in)=out is (semi)det
 */
:- type actionTable(State) == (func(State,token_kind) = parserAction).

/**
 * A state transition table for LR(k) parsers called goto table.
 * The goto table associates with each state q and transition symbol y
 * the state to which the parser (or its deterministic characteristic finite automaton
 * LR-DFA(G)) should make a transition. In the following a denotes the lookahead terminals.
 * In case of actionTable(q,a)=shift, gotoTable(q,tok) determines the transition from
 * the current state q under the token tok read (when pushing the target state).
 * In case of actionTable(q,a)=reduce(X&rarr;alpha), gotoTable(q,X) determines the
 * transition under X=y from the state that remains on the stack after having popped
 * the length(alpha) elements belonging to alpha from the stack.
 * @preconditions q:State & y:symbol
 *  & mode actionTable(in,in)=out is (semi)det
 */
:- type gotoTable(State) == (func(State,symbol) = State).

/**
 * Create the LR(1) parser.
 * @param grammar the grammar to parse.
 * @param initialState the initial state of the LR automaton to start parsing.
 * @param actionTable the action table of the LR(1) parser determining the parser action
 *  under some lookahead at a particular state.
 * @param gotoTable the goto table of the LR(1) parser determining the transition under
 *  some symbol in a particular state.
 */
:- all [State] some [P] (func lr1_parser(grammar/*(actionType)*/, State, actionTable(State), gotoTable(State)) = P => parser(P))
	<= printable(State).
:- mode lr1_parser(in,in,in(func(in, in) = out is det),in(func(in, in) = out is det)) = out is det.

%% specific version of lr1_parser for pragma memo to work
:- type lr1Parser(State).
:- instance printable(lr1Parser(State)) <= printable(State).
:- instance parser(lr1Parser(State)) <= printable(State).
:- all [State] (func lr1_parser_specific(grammar/*(actionType)*/, State, actionTable(State), gotoTable(State)) = lr1Parser(State))
	<= printable(State).
:- mode lr1_parser_specific(in,in,in(func(in, in) = out is det),in(func(in, in) = out is det)) = out is det.



:- implementation.

:- import_module list, require, std_util, string, basecap, scanners, stack, functionals, int.

:- type lr1Parser(State)
	---> lr1Parser(
		      grammar :: grammar/*(actionType)*/,
		      initialState :: State,
		      actionTable :: actionTable(State),
		      gotoTable :: gotoTable(State)
		     ).
:- instance printable(lr1Parser(State)) <= printable(State) where [
	to_string(lr1Parser(G, I0, AT, GT)) = "LR(1)[" ++ to_string(G)
	    ++ ": initialState=" ++ to_string(I0)
	    ++ ", actionTable="
	    /*++ to_string(solutions_set((pred(("(" ++ to_string(S) ++ to_string(T) ++ "):" ++ to_string(A))::out) is nondet :-
			     AT(S,T)=A)))*/
	    ++ ", gotoTable="
	    /*++ to_string(solutions_set((pred(("(" ++ to_string(S) ++ to_string(X) ++ "):" ++ to_string(Sp))::out) is nondet :-
			     GT(S,X)=Sp)))*/
	    ++ "]"
      ].
:- type stack_symbol(State)/*(actionType)*/
 	---> stack_symbol(
 			  /**
 			   * The parser state that has been pushed to the stack.
 			   * Comes from a finite set.
 			   */
 			  state :: State,
 			  /**
 			   * The value that has been associated with this stack
 			   * entry during evaluation.
 			   */
 			  stack_value :: actionType
 			 ).
:- instance printable(stack_symbol(State)) <= printable(State) where [
	to_string(stack_symbol(S,A)) = to_string(S)
 ].

:- instance printable(parserAction) where [
	to_string(shift) = "shift",
	to_string(reduce(P)) = "reduce(" ++ to_string(P) ++ ")",
	to_string(accept) = "accept",
	to_string(error) = "error"
 ].

parserAction_equals(shift,shift).
parserAction_equals(reduce(rule(X,Alpha,_)),reduce(rule(X,Alpha,_))).
parserAction_equals(accept,accept).
parserAction_equals(error,error).

/**
 * reducingParse(Grammar, Tokenstream, WorkingStack, AST)
 * Predict next production.
 * LR(k)-parser
 * <pre>
 * action: Q x T<sup>&le;k</sup> to {shift,reduce,accept,error}
 * goto: Q x V to state
 * lookahead: symbol
 * S: stack of state
 * scan k
 * push(S, q0)
 * forever do
 *   case action(top(S), lookahead) of
 *    shift: push(S, goto(top(S), head(lookahead)));
 *           lookahead = tail(lookahead);
 *   	     scan
 *    reduce(X->A): pop(S) length(A) times;
 *           push(S, goto(top(S),X));
 *   	     output(X->A);
 *    accept: stop
 *    error: error("Syntax error")
 *   end case
 * od
 * </pre>
 */
:- pred reducingParse(lr1Parser(State)::in, TS::in, stack(stack_symbol(State))::in, actionType::out) is semidet
	<= (tokenSequence(TS), printable(State)).


:- instance parser(lr1Parser(State)) <= printable(State) where [
	parse(P, TS, Output) :-
		reducingParse(P, TS, push(init, stack_symbol(initialState(P), new_action_value("UNDEFINED"))), Output)
  ].
lr1_parser(Grammar,InitialState,ActionTable,GotoTable) = lr1Parser(Grammar,InitialState,ActionTable,GotoTable).
lr1_parser_specific(Grammar,InitialState,ActionTable,GotoTable) = R :-
	R = lr1Parser(Grammar,InitialState,ActionTable,GotoTable).


/**
 * pop(Stack, ExpectedSymbols, ValuesFound, RemainingStack)
 */
:- pred popExtract(stack(stack_symbol(State)/*(actionType)*/), int, list(actionType), stack(stack_symbol(State)/*(actionType)*/)).
:- mode popExtract(in,in,out,out) is semidet.
popExtract(Stack, N, ValuesFound, Stack2) :-
        require(N >= 0, "non-negative nesting expected"),
	( N = 0 ->
	    ValuesFound = [],
	    Stack2 = Stack
	;
	    N > 0,
	    pop(Stack, stack_symbol(_, V), Stack_p),
	    ValuesFound = [V|ValuesFound2],
	    popExtract(Stack_p, N-1, ValuesFound2, Stack2)
	).

reducingParse(Parser, Tokens, Stack, Output) :-
	get_token(LOOKAHEAD, Tokens, TokensR),
	ParserAction = actionTable(Parser)(state(top_det(Stack)), get_kind(LOOKAHEAD)),
	(
	  ParserAction = shift,
	  Stack2 = push(Stack,
			stack_symbol(
				     gotoTable(Parser)(state(top_det(Stack)), 'new terminal'(get_kind(LOOKAHEAD))),
				     new_action_value(get_value(LOOKAHEAD))
				    )
		       ),
	  %%log_print("  " ++ to_string(ParserAction) ++ " from " ++ to_string(Stack) ++ "\n\tleaving " ++ to_string(Stack2)),
	  reducingParse(Parser, TokensR, Stack2, Output)
	;
	  ParserAction = reduce(Prod),
	  %% Stack_p = nest(pop2, length(rhs(Prod)))(Stack),
	  popExtract(Stack, length(rhs(Prod)), Values, Stack_p),
	  Value = action(Prod)(list__reverse(Values)),
	  Stack2 = push(Stack_p,
			stack_symbol(
				     gotoTable(Parser)(state(top_det(Stack_p)), nt(Prod)),
				     Value
				    )
		       ),
	  %%log_print("  " ++ to_string(ParserAction) ++ " from " ++ to_string(Stack) ++ "\n\tleaving " ++ to_string(Stack2)),
	  reducingParse(Parser, Tokens, Stack2, Output)
	;
	  ParserAction = accept,
	  %%log_print("  " ++ to_string(ParserAction) ++ " from " ++ to_string(Stack)),
	  Output = stack_value(top_det(Stack))
	;
	  ParserAction = error,
	  error("Syntax error")
	).

:- end_module parsers__lr1Parser.