/**
 * Provides a deterministic LR(1) bottom-up parser implementation.
 * Produces right derivations.
 * @author Andre Platzer
 * @version 2003-08-05
 */
:- module parsers__lrParser.
:- interface.

:- type parserAction
	---> shift
	;    reduce(production)
	;    accept
	;    error.
/**
 * An action table that associates with each state q and lookahead terminals a
 * the action that the parser should do.
 * @preconditions q:State & a:list(token)
 */
:- type actionTable(State) == (func(State,token) = parserAction).
/**
 * A goto table that associates with each state q and transition symbol y
 * the state to which the parser (or its deterministic characteristic finite automaton
 * LR-DFA(G)) should make a transition. In the following a denotes the lookahead terminals.
 * In case of actionTable(q,a)=shift, gotoTable(q,tok) determines the transition from
 * the current state q under the token tok read (when pushing the target state).
 * In case of actionTable(q,a)=reduce(X&rarr;alpha), gotoTable(q,X) determines the
 * transition under X=y from the state that remains on the stack after having popped
 * the length(alpha) elements belonging to alpha from the stack.
 * @preconditions q:State & y:symbol
 */
:- type gotoTable(State) == (func(State,symbol) = State).

/**
 * Create the LR(1) parser.
 */
:- all [State] some [P] (func lr_parser(grammar/*(actionType)*/, State, actionTable(State), gotoTable(State)) = P => parser(P))
	<= printable(State).
:- mode lr_parser(in,in,in(func(in, in) = out is det),in(func(in, in) = out is det)) = out is det.


:- implementation.

:- import_module list, require, std_util, string, basecap, scanners, stack, functionals.

:- type lrParser(State)
	---> lrParser(
		      grammar :: grammar/*(actionType)*/,
		      initialState :: State,
		      actionTable :: actionTable(State),
		      gotoTable :: gotoTable(State)
		     ).
%%:- type stack_symbol(State) == State.

/**
 * reducingParse(Grammar, Tokenstream, WorkingStack, AST)
 * Predict next production.
 * LR(k)-parser
 * <pre>
 * action: Q x T<sup>&le;k</sup> to {shift,reduce,accept,error}
 * goto: Q x V to state
 * lookahead: symbol
 * S: stack of state
 * scan k
 * push(S, q0)
 * forever do
 *   case action(top(S), lookahead) of
 *    shift: push(S, goto(top(S), head(lookahead)));
 *           lookahead = tail(lookahead);
 *   	     scan
 *    reduce(X->A): pop(S) length(A) times;
 *           push(S, goto(top(S),X));
 *   	     output(X->A);
 *    accept: stop
 *    error: error("Syntax error")
 *   end case
 * od
 * </pre>
 */
:- pred reducingParse(lrParser(State)::in, TS::in, stack(/*stack_symbol*/State)::in, actionType::out) is semidet
	<= (tokenSequence(TS), printable(State)).


:- instance parser(lrParser(State)) <= printable(State) where [
	parse(P, TS, Output) :-
		reducingParse(P, TS, push(init, initialState(P)), Output)
  ].
lr_parser(Grammar,InitialState,ActionTable,GotoTable) = lrParser(Grammar,InitialState,ActionTable,GotoTable).


:- func pop2(stack(T)) = stack(T).
:- mode pop2(in) = out is det.
pop2(Stack) = Stack2 :-
	pop_det(Stack,_,Stack2)
	/*;
	error("Empty stack could not pop2. Declaring semidet would need multiple mode higher-order function nest")*/.
	 
reducingParse(Parser, Tokens, Stack, Output) :-
	get_token(LOOKAHEAD, Tokens, TokensR),
	ParserAction = actionTable(Parser)(top_det(Stack),LOOKAHEAD),
	(
	  ParserAction = shift,
	  Stack2 = push(Stack, gotoTable(Parser)(top_det(Stack),LOOKAHEAD)),
	  reducingParse(Parser, TokensR, Stack2, Output)
	;
	  ParserAction = reduce(Prod),
	  Stack_p = nest(pop2, length(rhs(Prod)))(Stack),
	  Stack2 = push(Stack_p, gotoTable(Parser)(top_det(Stack_p),nt(Prod))),
	  Output = to_string(Prod) ++ "\n  " ++ Output3,
	  Output3= "stack2=" ++ to_string(Stack2) ++ " from stack=" ++ to_string(Stack) ++ " tokens=" ++ to_string(Tokens) ++ "\n  " ++ Output2,
	  reducingParse(Parser, Tokens, Stack2, Output2)
	;
	  ParserAction = accept,
	  Output = "Accept stack=" ++ to_string(Stack) ++ " tokens=" ++ to_string(Tokens)
	;
	  ParserAction = error,
	  error("Syntax error")
	).
