/**
 * Provides abstract parser data representations.
 * @author Andre Platzer
 * @version 2003-08-01
 */
:- module parsers.
:- interface.
:- include_module syntheticParser, sllParser, analyticParser, lr1Parser.
:- import_module scanners, basecap, list, set.

/**
 * The types of lexical symbol or token representations.
 * @author Andre Platzer
 * @version 2003-08-01
 */
:- type symbol
	%% terminal lexical symbols
	---> some [TokenType] (terminal(
		      TokenType
		     )
		=> token(TokenType))
	%% grammatical symbols
	/**
	 * nonterminal parser symbols
	 */
	;    nonterminal(
			 name :: string
			).

:- instance printable(symbol).
:- type symbol_chain == list(symbol).

/**
 * Create new terminal.
 * @note Generally preferred to the type constructor terminal/1.
 */
:- all [TokenType] func new_terminal(TokenType) = symbol
	<= token(TokenType).
/**
 * Create new nonterminal.
 */
:- func new_nonterminal(string) = symbol.

/**
 * Checks whether a symbol is a terminal.
 */
:- pred is_terminal(symbol).
:- mode is_terminal(in) is semidet.
%%:- promise all [O] (is_terminal(O) <=> (some [A] (O = terminal(A)))).

/**
 * Checks whether a symbol is a nonterminal.
 */
:- pred is_nonterminal(symbol).
:- mode is_nonterminal(in) is semidet.
%%:- promise all [O] (is_nonterminal(O) <=> (some [N] (O = nonterminal(N)))).


/**
 * Production rule representation.
 * rule(nonterminal(X), alpha, action) represents the context-free CH-2 production
 * <center>X&rarr;alpha</center>
 * with the given semantic action to happen on each application of the production.
 * When having applied the production
 * X&rarr;Y1 Y2 ... Yn
 * the parser also invokes X.value = action([Y1.value,Y2.value,...,Vn.value])
 * as a synthesized LAG(1) evaluation.
 * @param Action the type of action values.
 * @author Andre Platzer
 * @version 2003-08-01
 */
:- type production/*(actionType)*/
	---> rule(
		  nt :: symbol,
		  rhs :: list(symbol),
		  action :: func(list(actionType)) = actionType
		 %% @todo would like to declare "is semidet" via some inst or mode or type
		 )
	.%%% where equality is production_equals.
:- instance printable(production/*(actionType)*/).

/**
 * Create new productions.
 * @note Generally preferred to the type constructor rule/3.
 */
:- func new_rule(symbol, list(symbol), (func(list(actionType))=actionType)) = production.
:- mode new_rule(in, in, in(func(in)=out is det)) = out is det.
:- mode new_rule(in, in, in(func(in)=out is semidet)) = out is det.

:- pred production_equals(production,production).
:- mode production_equals(in,in) is semidet.
:- mode production_equals(in,out) is det.
:- mode production_equals(out,in) is det.

/**
 * Get the nonterminals of a production.
 * @param production X0&rarr;u0 X1 u1 X2 u2 ... Xn un
  
 * @return the projection to nonterminals [X0,X1,X2,...,Xn].
 */
:- func production_nonterminals(production/*(actionType)*/) = list(symbol).
:- mode production_nonterminals(in) = out is det.

/**
 * Get the i-th nonterminal of a production.
 * @param production X0&rarr;u0 X1 u1 X2 u2 ... Xn un
 * @param i the index of the nonterminal to get (0 is left hand side).
 * @return Xi.
 */
:- func production_nonterminal(production/*(actionType)*/,int) = symbol.
:- mode production_nonterminal(in,in) = out is semidet.


/**
 * Represents a context-free CH-2 grammar.
 * @param Action the type of action values.
 * @author Andre Platzer
 * @version 2003-08-01
 */
:- type grammar/*(actionType)*/
	---> grammar(
		     %% implicit terminals :: symbols,
		     %% implicit nonterminals :: symbols,
		     /**
		      * The starting symbol of this grammar, where all derivations start.
		      */
		     start :: symbol,
		     productions :: list(production/*(actionType)*/)
		    ).
:- instance printable(grammar).


/**
 * A parser typeclass.
 * @param Parser the concrete parser type.
 * @param Action the type of action values.
 * @author Andre Platzer
 * @version 2003-08-04
 * @todo distinguish TS containing only terminals, from stack, containing terminals or other symbols.
 */
:- typeclass parser(Parser) <= printable(Parser) where [
	/**
	  * parse(Parser, Input, AST)
	  * Let Parser parse the token sequence producing the output AST.
	  * @param AST the resulting final action value of the top-most occurrence
	  *  of a nonterminal, the starting symbol.
	  * @todo generalize output to whatever definite clause grammars produce.
	  */
	pred parse(Parser, TS, actionType) <= tokenSequence(TS),
	mode parse(in, in, out) is nondet
].

/**
 * Determines the vocabulary V=N union T used in a grammar G.
 */
:- func vocabulary(grammar) = set(symbol).
:- mode vocabulary(in) = out is det.

/**
 * Determines the nonterminals N used in a grammar G.
 */
:- func nonterminals(grammar) = set(symbol).
:- mode nonterminals(in) = out is det.

/**
 * Determines the terminals T used in a grammar G.
 */
:- func terminals(grammar) = set(symbol).
:- mode terminals(in) = out is det.


/**
 * The value type for production actions.
 * @see std_util#univ
 * @version 2003-08-06
 */
:- type actionType.
:- instance printable(actionType).
/**
 * Create a new action value representation from any type.
 */
:- all [ActionValueType] (func new_action_value(ActionValueType) = actionType
			 <= printable(ActionValueType)).
:- mode new_action_value(in) = out is det.

/**
 * Get the action value content of some type.
 */
:- some [ActionValue] (func action_value(actionType) = ActionValue
%%			 => printable(ActionValue)
		      ).
:- mode action_value(in) = out is det.

/**
 * Try to get the action value content of a particular type.
 * Attempt fails if the types are not compatible.
 * Can be used for casting like in
 * Value = action_value_to_type(A) `with_type` syntaxTree(string)
 */
:- all [ActionValue] (func action_value_to_type(actionType) = ActionValue).
:- mode action_value_to_type(in) = out is semidet.



:- implementation.
:- import_module string, require, std_util.

new_terminal(Tok) = 'new terminal'(Tok).
new_nonterminal(X) = nonterminal(X).

is_terminal(terminal(_)).

is_nonterminal(nonterminal(_)).

:- instance printable(symbol) where [
        to_string(terminal(A)) = to_string(A),
        to_string(nonterminal(A)) = "'" ++ A ++ "'"
].

:- pragma promise_pure(new_rule/3).
new_rule(X::in, Alpha::in, Action::in(func(in)=out is det))
 = (rule(X,Alpha,Action)::out).
new_rule(X::in, Alpha::in, Action::in(func(in)=out is semidet))
 = (rule(X,Alpha,
	semidet_to_det(Action,
		       "action for " ++ to_string(rule(X,Alpha,det_head)))
       )::out).


/**
 * @internal Currently we use duplicate storage. Once as a univ in order to
 *  have univ_to_type for semideterministically casting it to all types.
 *  And second as an existential type of type class constraint printable,
 *  since according to "Some idioms using existentially quantified types"
 *  "Using `univ' doesn't work if you also want to use type class constraints".
 *  So we simply store the same value twice.
 */
:- type actionType
	---> some [ActionValueType1] (actionType(
			univ,
			ActionValueType1
		       )
				     => printable(ActionValueType1)
				     ).
:- instance printable(actionType) where [
	to_string(actionType(_,V)) = to_string(V)
  ].

new_action_value(X) = 'new actionType'(univ(X),X).
action_value(actionType(X,_)) = univ_value(X).
action_value_to_type(actionType(X,_)) = Y :-
	univ_to_type(X, Y).


:- func prettyPrint(list(symbol)) = string.
prettyPrint([]) = "".
prettyPrint([X]) = to_string(X).
prettyPrint([X1,X2|Xs]) = to_string(X1) ++ " " ++ prettyPrint([X2|Xs]).

:- instance printable(production/*(actionType)*/) where [
	to_string(rule(A, ALPHA, COMP)) = to_string(A) ++ " -> " ++ prettyPrint(ALPHA)
  ].

:- instance printable(grammar) where [
	to_string(grammar(S,Pi)) = "(" ++ to_string(S) ++ "," ++ to_string(Pi) ++ ")"
  ].


:- pragma promise_pure(production_equals/2).
production_equals(rule(X,Alpha,_)::in, rule(X,Alpha,_)::in).
production_equals(rule(X,Alpha,COMP)::in, rule(X,Alpha,COMP)::out).
production_equals(rule(X,Alpha,COMP)::out, rule(X,Alpha,COMP)::in).

production_nonterminals(rule(X,Alpha,_)) = filter(pred(nonterminal(_)::in) is semidet, [X|Alpha]).

%% @internal could be optimized
production_nonterminal(rule(X, Alpha, COMP), I)
 = index1_det(production_nonterminals(rule(X, Alpha, COMP)), I).


vocabulary(G) = list_to_set(condense(map(
					 (func(rule(X,Alpha,_)) = [X|Alpha]),
					 productions(G)
					))).

nonterminals(G) = filter(is_nonterminal, vocabulary(G)).
terminals(G) = filter(is_terminal, vocabulary(G)).


:- end_module parsers.