/**
 * Provides a deterministic SLL(k) top-down parser implementation.
 * Especially notice SLL(1)=LL(1)
 * @author Andre Platzer
 * @version 2003-08-01
 * @structure specializes parsers.syntheticParser
 */
:- module parsers__sllParser.
:- interface.
:- import_module scanners.

/**
 * Predicts the production to use for a nonterminal under some lookahead terminals.
 * parseTable(X,a)=alpha specifies that the production to apply for the
 * nonterminal X at a position with lookahead a is
 * X->alpha.
 * @preconditions a:word & X:nonterminal. 
 * @postconditions alpha:symbol_chain & member(rule(X,alpha),grammar)
 *  & mode parseTable(in,in)=out is (semi)det 
 */
:- type parseTable == (func(symbol,word_kind) = symbol_chain).

/**
 * Create the SLL(k) parser.
 * @param grammar the grammar to parse.
 * @param k the length of the lookahead to use.
 * @param parseTable the table predicting the production to use under some lookahead.
 * @xxx still only works for k=1
 * @internal still ignores evaluation actions.
 */
:- some [P] (func sll_parser(grammar/*(actionType)*/, int, parseTable) = P => parser(P)).
:- mode sll_parser(in,in,in(func(in,in)=out is det)) = out is det.
:- mode sll_parser(in,in,in(func(in,in)=out is semidet)) = out is det.


:- implementation.

:- import_module require, exception, std_util, string, basecap.

:- type sllParser/*(actionType)*/
	---> sllParser(
		       grammar :: grammar/*(actionType)*/,
		       lookahead :: int,
		       parsetable :: parseTable
		     ).
:- instance printable(sllParser) where [
	to_string(sllParser(G, K, PT)) = "SLL(" ++ to_string(K) ++ ")[" ++ to_string(G)
	    ++ ", parseTable="
	    /*++ to_string(solutions_set((pred(("(" ++ to_string(S) ++ to_string(T) ++ "):" ++ to_string(A))::out) is nondet :-
			     AT(S,T)=A)))*/
	    ++ "]"
      ].

:- type stack_symbol == symbol.

/**
 * predictiveParse(Grammar, Tokenstream, WorkingStack, AST)
 * Predict next production.
 */
:- pred predictiveParse(sllParser/*(actionType)*/::in, TS::in, list(stack_symbol)::in, actionType::out)
	is semidet
	<= tokenSequence(TS).


:- instance parser(sllParser/*,actionType*/) where [
	parse(P, TS, Output) :-
		predictiveParse(P, TS, [start(grammar(P))], Output)
  ].
sll_parser(Grammar, K, PT) = sllParser(Grammar, K, PT).


/**
 * accept
 */
predictiveParse(_, Tokens, [], new_action_value("")) :-
	empty(Tokens).

/**
 * expand nonterminal alternative.
 */
predictiveParse(Parser, Tokens, [nonterminal(X) | Stack], Output) :-
	%%%member(rule(nonterminal(X), Righthandside), productions(grammar(Parser))),
	get_token(Tok, Tokens, _),
	%% @todo optimize this should be implemented as a table lookup in Tok,nonterminal(X). Mercury does indexing, doesn't it?
	(
	  Righthandside = parsetable(Parser)(nonterminal(X), [get_kind(Tok)])
	;
	  throw("Parse exception: for " ++ to_string(nonterminal(X)) ++ ", encountered " ++ to_string(Tok))
	),
	append(Righthandside, Stack, Stack2),
	predictiveParse(Parser, Tokens, Stack2, Output2),
	Output = Output2.
	%%Output = new_action_value(to_string(rule(nonterminal(X), Righthandside)) ++ "\n  " ++ to_string(Output2)).

/**
 * compare terminal input.
 */
predictiveParse(Parser, Tokens, [terminal(A1) | Stack], Output) :-
	get_token(A2, Tokens, TokensR),
	token_class_equals(A1, A2),
	predictiveParse(Parser, TokensR, Stack, Output).

:- end_module parsers__sllParser.
