/**
 * Provides a general nondeterministic synthetic top-down parser implementation.
 * Produces left derivations.
 * @author Andre Platzer
 * @version 2003-08-01
 */
:- module parsers__syntheticParser.
:- interface.

/**
 * Create the nondeterministic synthetic top-down parser.
 */
:- some [P] (func synthetic_parser(grammar/*(actionType)*/) = P => parser(P)).
:- mode synthetic_parser(in) = out is det.


:- implementation.

:- import_module list, require, std_util, string, basecap, scanners.

:- type syntheticParser%%(actionType)
	---> syntheticParser(
			     grammar :: grammar%%(actionType)
			    ).
:- instance printable(syntheticParser) where [
	to_string(syntheticParser(G)) = "nondeterministic synthetic top-down parser[" ++ to_string(G) ++ "]"
      ].

:- type stack_symbol == symbol.

/**
 * predictiveParse(Grammar, Tokenstream, WorkingStack, AST)
 * Predict next production.
 * @internal still ignores evaluation actions.
 */
:- pred predictiveParse(syntheticParser/*(actionType)*/::in, TS::in, list(stack_symbol)::in, actionType::out) is nondet
	<= tokenSequence(TS).


:- instance parser(syntheticParser/*(actionType),actionType*/) where [
	parse(P, TS, Output) :-
		predictiveParse(P, TS, [start(grammar(P))], Output)
  ].
synthetic_parser(Grammar) = syntheticParser(Grammar).


/**
 * accept empty.
 * q# ---> q
 */
predictiveParse(_, Tokens, [], new_action_value("")) :-
	empty(Tokens).

/**
 * expand or generate nonterminal alternative.
 * Already reducing nonterminal(X) right ahead, presuming the production choice will prove right.
 * Xq ---> Rq  whenever  (X--->converse(R)) &isin; &Pi;
 */
predictiveParse(Parser, Tokens, [nonterminal(X) | Stack], Output) :-
	member(rule(nonterminal(X), Righthandside, COMP), productions(grammar(Parser))),
	append(Righthandside, Stack, Stack2),
	predictiveParse(Parser, Tokens, Stack2, Output2),
	Output = new_action_value(to_string(rule(nonterminal(X), Righthandside, COMP)) ++ "\n  " ++ to_string(Output2)).

/**
 * compare terminal input with stack (corresponds to predictive shift).
 * aqa ---> q
 */
predictiveParse(Parser, Tokens, [terminal(A1) | Stack], Output) :-
	get_token(A2, Tokens, TokensR),
	token_class_equals(A1, A2),
	predictiveParse(Parser, TokensR, Stack, Output).

/**
 * Error handler.
 */
/*predictiveParse(_, Tokens, [Delta | _], _) :-
	get_token(Tok, Tokens, _),
	not Tok = Delta,
	error("Syntactic error. expected: " ++ to_string(Delta) ++ ", found: " ++ to_string(Tok)).
*/
/**
 * Cannot happen.
 */
/*predictiveParse(_, Tokens, Stack, Output) :-
	error("impossible things should never have happened. Incomplete case distinction in error handler clause predictiveParse/4. "
	     ++ " Tokens=" ++ to_string(Tokens)
	     ++ ", Stack=" ++ to_string(Stack)).
*/
