/**
 * Provides a tokenizer for regular expressions.
 * @author Andre Platzer
 * @version 2003-08-21
 */
:- module regexp_tokenizer.
:- interface.
:- import_module scanners, char, basecap.

:- type regexp_scanner.
:- instance printable(regexp_scanner).
:- instance scanner(regexp_scanner).

/**
 * Create a tokenizer for regular expressions.
 */
%%:- some [S] (func regexp_tokenizer = S => scanner(S)).
:- func regexp_tokenizer = regexp_scanner.
:- mode regexp_tokenizer = out is det.

:- type regexp_token
	---> character(
		  signifier :: char
		 )
	;    alternation
	;    iteration
	;    begin_group
	;    end_group
	;    range
	;    open
	;    close
	;    position
	;    endoffile.
:- instance printable(regexp_token).
:- instance token(regexp_token).

/**
 * Get token kind common for all character(A).
 * Convenience.
 */
:- func character = token_kind.



:- implementation.

:- import_module list, string, require, std_util.
:- import_module scanner_util, string_source.


:- all [TSIn] some[TS] (func scan(TSIn) = TS => tokenSequence(TS))
	<= tokenSequence(TSIn).
:- mode scan(in) = out is semidet.


:- type regexp_scanner
	---> regexp_scanner.

:- instance printable(regexp_scanner) where [
	to_string(_) = "regexp_scanner tokenizes regular expressions"
  ].
:- instance scanner(regexp_scanner) where [
	scan(S, Input) = scan(Input)
  ].

regexp_tokenizer = regexp_scanner.


:- type token == regexp_token.

:- instance printable(regexp_token) where [
        to_string(character(A)) = char_to_string(A),
        to_string(alternation) = "|",
        to_string(iteration) = "*",
        to_string(open) = "(",
        to_string(close) = ")",
        to_string(begin_group) = "[",
        to_string(end_group) = "]",
        to_string(range) = "-",
        to_string(position) = ".",
        to_string(endoffile) = "end_of_file"
].

:- instance token(regexp_token) where [
	(get_value(T) = to_string(T) :-
	    T = end_of_file ->
	        error("cannot get_value of " ++ to_string(endoffile))
	    ;
	        true
	),

	(get_kind(T) = make_token_kind(Kind) :-
	    T = character(_) ->
	        Kind = "character"
	    ;
	        Kind = to_string(T)
	),

	end_of_file = endoffile
 ].


scan(Input) = list__reverse(Tokens) :-
	scan(Input, [], Tokens).
:- pred scan(TSIn, list(token), list(token)) <= tokenSequence(TSIn).
:- mode scan(in, in, out) is det.
scan(Input, Toks0, Toks) :-
	get_token(Char1, Input, InputRest),
	Char = cast_det(Char1, "character-based tokenSequence expected"),
	(
	  Char = end_of_file_source,
	  Toks = [endoffile | Toks0]
	;
	  Char = char_token(C),
	  (
	    %% @xxx should remove this for specifying skip tokens?
	    char__is_whitespace(C)
	  ->
	    scan(InputRest, Toks0, Toks)
	  ;
	    (C = '|')
	  ->
	    scan(InputRest, [alternation | Toks0], Toks)
	  ;
	    (C = (*))
	  ->
	    scan(InputRest, [iteration | Toks0], Toks)
	  ;
	    (C = '(')
	  ->
	    scan(InputRest, [open | Toks0], Toks)
	  ;
	    (C = ')')
	  ->
	    scan(InputRest, [close | Toks0], Toks)
	  ;
	    (C = '[')
	  ->
	    scan(InputRest, [begin_group | Toks0], Toks)
	  ;
	    (C = ']')
	  ->
	    scan(InputRest, [end_group | Toks0], Toks)
	  ;
	    (C = ('-'))
	  ->
	    scan(InputRest, [range | Toks0], Toks)

	  ;
	    (C = ('.'))
	  ->
	    scan(InputRest, [position | Toks0], Toks)
	  ;
	    scan(InputRest, [character(C) | Toks0], Toks)
	  )
	).

character = get_kind(character('x')).


:- end_module regexp_tokenizer.