/**
 * Regular expression parsing with item transition.
 * @author Andre Platzer
 * @version 2003-08-21
 */
:- module regexp_transition.
:- interface.
:- import_module string, set.
:- use_module char_group.

	/**
	 * Makes an epsilon transition of the regular expression item.
	 * @return the set of possible successor regular expressions
	 * of the given regular expression reachable without consuming
	 * any input.
	 */
	:- func regexp_epsilon_transition(string) = set(string).
	:- mode regexp_epsilon_transition(in) = out is semidet.

	/**
	 * Get the character groups appearing in a regular expression.
	 * A single character c counts as a character group [c]
	 */
	:- func regexp_char_groups(string) = set(char_group__char_group).
	:- mode regexp_char_groups(in) = out is semidet.



:- implementation.
:- import_module string, list, int, require, std_util, basecap, scanners, parsers,
	parsers__lr1Parser,
	parser_generator, parser_generator__lr1Generator, parser_generator__sllGenerator,
	regexp_tokenizer, string_source.

:- (func regexp_epsilon_transition = lr1Parser(int)).
:- pragma memo(regexp_epsilon_transition/0).
regexp_epsilon_transition = P :-
	AllLinearizations = alllinearizations,
	%% the regular expression item grammar along with actions that
	%% make an epsilon transition
	RIG = grammar(nonterminal("R"),
	[
	 /* @see regexpi.grammar */
	 new_rule(nonterminal("R"), [nonterminal("R"), nonterminal("Rc")],
		  AllLinearizations
	     ),
	 new_rule(nonterminal("R"), [nonterminal("Rc")],
		  AllLinearizations
	     ),
	 new_rule(nonterminal("Rc"), [new_terminal(open), nonterminal("R"), new_terminal(alternation), nonterminal("R"), new_terminal(close)],
		  AllLinearizations
	     ),
	 new_rule(nonterminal("Rc"), [new_terminal(open), nonterminal("R"), new_terminal(close), new_terminal(iteration)],
		  AllLinearizations
	     ),
	 new_rule(nonterminal("Rc"), [new_terminal(character)],
		  AllLinearizations
	     ),
	 new_rule(nonterminal("Rc"), [new_terminal(begin_group), new_terminal(character), new_terminal(range), new_terminal(character), new_terminal(end_group)],
		  AllLinearizations
	     ),
	 %% @xxx we do not yet have a transition rule for this  .(R) ==> (.R) and (R.) ==> (R)., since this rules is essentially unnecessary in our regular expressions
	 new_rule(nonterminal("Rc"), [new_terminal(open), nonterminal("R"), new_terminal(close)],
		  AllLinearizations
	     ),

	 %% epsilon transition transformation cases
	 new_rule(nonterminal("Rc"), [new_terminal(position), new_terminal(open), nonterminal("R"), new_terminal(close), new_terminal(iteration)],
		  (func([_,_,R,_,_]::in)
		    = (new_action_value(relinearize(expand([single("("),single("."),action_value_to_type(R),single(")"),single("*")]))
				       `union`
				       relinearize(expand([single("("),action_value_to_type(R),single(")"),single("*"),single(".")]))
				       )::out) is semidet)
	     ),
	 new_rule(nonterminal("Rc"), [new_terminal(open), nonterminal("R"), new_terminal(position), new_terminal(close), new_terminal(iteration)],
		  (func([_,R,_,_,_]::in)
		    = (new_action_value(relinearize(expand([single("("),single("."),action_value_to_type(R),single(")"),single("*")]))
				       `union`
				       relinearize(expand([single("("),action_value_to_type(R),single(")"),single("*"),single(".")]))
				       )::out) is semidet)
	     ),

	 new_rule(nonterminal("Rc"), [new_terminal(position), new_terminal(open), nonterminal("R"), new_terminal(alternation), nonterminal("R"), new_terminal(close)],
		  (func([_,_,R,_,S,_]::in)
		    = (new_action_value(relinearize(expand([single("("),single("."),action_value_to_type(R),single("|"),action_value_to_type(S),single(")")]))
				       `union`
				       relinearize(expand([single("("),action_value_to_type(R),single("|"),single("."),action_value_to_type(S),single(")")]))
				       )::out) is semidet)
	     ),
	 new_rule(nonterminal("Rc"), [new_terminal(open), nonterminal("R"), new_terminal(position), new_terminal(alternation), nonterminal("R"), new_terminal(close)],
		  (func([_,R,_,_,S,_]::in)
		    = (new_action_value(relinearize(expand([single("("),action_value_to_type(R),single("|"),action_value_to_type(S),single(")"),single(".")]))
				       )::out) is semidet)
	     ),
	 new_rule(nonterminal("Rc"), [new_terminal(open), nonterminal("R"), new_terminal(alternation), nonterminal("R"), new_terminal(position), new_terminal(close)],
		  (func([_,R,_,S,_,_]::in)
		  = (new_action_value(relinearize(expand([single("("),action_value_to_type(R),single("|"),action_value_to_type(S),single(")"),single(".")]))
				     )::out) is semidet)
	     )
	]
	       ),
	log_print("\tregular expression transition grammar " ++ to_string(RIG)),
	(
	  P = lr1_parser_specific(RIG)
	;
	  error("internal error: regular expression item grammar")
	).

/**
 * Really only a secondary attribute to regexp_epsilon_transition with items removed.
 */
:- (func regexp_char_groups = lr1Parser(int)).
:- pragma memo(regexp_char_groups/0).
regexp_char_groups = P :-
	%% the regular expression item grammar along with actions that
	%% make an epsilon transition
	RG = grammar(nonterminal("R"),
	[
	 /* @see regexpi.grammar */
	 new_rule(nonterminal("R"), [nonterminal("R"), nonterminal("Rc")],
		  (func([C1,C2]::in)
		  = (new_action_value(
				      union(action_value_to_type(C1),action_value_to_type(C2))
				     `with_type` set(char_group__char_group)
				     )::out) is semidet)
	     ),
	 new_rule(nonterminal("R"), [nonterminal("Rc")],
		  det_head
	     ),
	 new_rule(nonterminal("Rc"), [new_terminal(open), nonterminal("R"), new_terminal(alternation), nonterminal("R"), new_terminal(close)],
		  (func([_,C1,_,C2,_]::in)
		  = (new_action_value(
				      union(action_value_to_type(C1),action_value_to_type(C2))
				     `with_type` set(char_group__char_group))::out) is semidet
		  )
	     ),
	 new_rule(nonterminal("Rc"), [new_terminal(open), nonterminal("R"), new_terminal(close), new_terminal(iteration)],
		  converse(index1_det,2)
	     ),
	 new_rule(nonterminal("Rc"), [new_terminal(character)],
		  (func([C]::in)
		  = (new_action_value(make_singleton_set(char_group__make_single_group(index_det(action_value_to_type(C),0))))::out) is semidet)
	     ),
	 new_rule(nonterminal("Rc"), [new_terminal(begin_group), new_terminal(character), new_terminal(range), new_terminal(character), new_terminal(end_group)],
		  (func([_,C1,_,C2,_]::in)
		  = (new_action_value(make_singleton_set(char_group__make_range_group(index_det(action_value_to_type(C1),0),index_det(action_value_to_type(C2),0))))::out) is semidet)
	     ),
	 new_rule(nonterminal("Rc"), [new_terminal(open), nonterminal("R"), new_terminal(close)],
		  converse(index1_det,2)
	     )
	]
	       ),
	log_print("\tregular expression character group grammar " ++ to_string(RG)),
	(
	  P = lr1_parser_specific(RG)
	;
	  error("internal error: regular expression grammar")
	).


regexp_epsilon_transition(Regexp) = Transitions :-
	Scanner = regexp_tokenizer,
	Parser = regexp_epsilon_transition,
	%%%log_print("\tregexp_trans " ++ Regexp ++ " scan"),
	TS = scan(Scanner, string_tokenSequence(Regexp)),
	( could_need_transition(Regexp) ->
	    log_print("\t?regexp_trans " ++ Regexp ++ " parse"),
	    Output = some_solution(parse(Parser, TS)),
	    log_print("\tRegexp transiton from " ++ Regexp ++ " to " ++ to_string(Output)),
	    Transitions = action_value_to_type(Output)
	;
	    Transitions = make_singleton_set(Regexp)
	).


regexp_char_groups(Regexp) = CharGroups :-
	Scanner = regexp_tokenizer,
	Parser = regexp_char_groups,
	log_print("\t?regexp_groups " ++ Regexp ++ " scan"),
	TS = scan(Scanner,string_tokenSequence(Regexp)),
	log_print("\t?regexp_groups " ++ Regexp ++ " parse"),
	(empty(TS) ->
	    CharGroups = set__init
	;
	    Output = some_solution(parse(Parser, TS)),
	    log_print("\tRegexp groups " ++ Regexp ++ " to " ++ to_string(Output)),
	    CharGroups = action_value_to_type(Output)
	).


%% nasty helpers

:- func single(string) = set(string).
single(S) = make_singleton_set(S).

/**
 * Whether the regular expression could perhaps make an epsilon transition.
 */
:- pred could_need_transition(string::in) is semidet.
could_need_transition(Regexp) :-
	PositionMarker = ('.'),
	Chars = to_char_list(Regexp),
	nth_member_search(Chars, PositionMarker, Position),
	index1(Chars, Position + 1, FollowsPosition),
	%% FollowsPosition is one of the critical transition characters?
	(
	  FollowsPosition = '('
	;
	  FollowsPosition = ')'
	;
	  FollowsPosition = '|'
	).


:- func extract_regexps(actionType) = set(string).
extract_regexps(V) = R :-
	(R2 = action_value_to_type(V) `with_type` set(string) ->
	    %% symbol case
	    R = R2
	; R3 = action_value_to_type(V) `with_type` string ->
	    %% token case
	    R = single(R3)
	;
	    report_lookup_error("unexpected action value type", action_value(V), V)
	).
	%%%log_print("\t\textract regexp " ++ to_string(R)).

:- func relinearize(set(list(string))) = set(string).
relinearize(S) = R :-
	R = map(string__append_list, S).
	%%%log_print("\t\trelinearize " ++ to_string(S) ++ " to\n\t\t\t" ++ to_string(R)).

:- func alllinearizations(list(actionType)) = actionType.
alllinearizations(List)
    = new_action_value(relinearize(
				expand(
				       list__map(extract_regexps,
						 List)
				      )
			       )
		      ).
	

/**
 * Distributive expand, makes all linearizations.
 * Transforms
 * [{a,b},{c,d}] to {[a,c],[a,d],[b,c],[b,d]}
 */
:- func expand(list(set(T))) = set(list(T)).
expand(List) = foldr(func(Prefixes,Suffixes)
		      = fold(func(Prefix,Concatenations)
			      = Concatenations `union` map(append([Prefix]), Suffixes),
			     Prefixes,
			     set__init
			     ),
		     List,
		     %% {&epsilon;}
		     make_singleton_set([])).

:- end_module regexp_transition.