/**
 * Provides scanner generator with LR(0) technique.
 * @author Andre Platzer
 * @version 2003-08-20
 */
:- module scanner_generator__lr0Generator.
:- interface.

:- import_module scanners, scanners__dfa, basecap.

/**
 * Generate scanner with LR(0) techniques.
 * Calculates the scanner parameters for the DFA scanner from an (appropriate) grammar.
 * @param grammar the grammar to parse.
 * @param transitionTable the transition table of deterministic finite automaton
 *  determining the transition under some symbol in a particular state.
 * @param acceptionMap the acception map of the deterministic finite acceptor
 *  determining which states are accepting states, and what the outputs tokens are.
 * @param initialState the initial state of the DFA to start parsing.
 * @see scanners__dfa#dfa_scanner/3
 */
:- all [Token] some [State,CharClass] (pred dfa_generate(grammar(Token),
					 transitionTable(State,CharClass),
					 acceptionMap(State,Token),
					 State,
					 charClassMapper(CharClass))
			    => (printable(State),printable(CharClass)))
        <= token(Token).
:- mode dfa_generate(in,out/*(transitionTable)*/,out,out,out) is semidet.

/**
 * Create a scanner based on a deterministic finite automaton.
 * @param grammar the lexical grammar to scan.
 * @see dfa_generate/4
 * @see scanners__dfa#dfa_scanner/3
 */
:- all [Token] some [S] (func dfa_scanner(grammar(Token)) = S
			=> scanner(S))
        <= token(Token).
:- mode dfa_scanner(in) = out is semidet.



:- implementation.

:- import_module scanners, basecap, list, set, map, string, int, std_util, functionals, require,
	dynatraf_set,
	regexp_transition.
:- import_module char_grouper, string_source.

%% map version of transitionTable
:- type transitionTableMap(State,CharClass) == map({State,CharClass}, State).

/**
 * A regular scanner item.
 * item(A,alpha,beta) is depicted [A->alpha.beta]
 * and represents the scanners state of having advanced to the position marked
 * by the point while parsing the production
 * A->alphabeta 
 */
:- type item
	---> item(
		  tok :: token_kind,
		  history :: list(atomic_regular_expression),
		  future :: list(atomic_regular_expression)
		 ).
:- instance printable(item) where [
	to_string(item(A,Alpha,Beta)) =
	    "[" ++ to_string(A) ++ " -> " ++ to_string(Alpha) ++ "." ++ to_string(Beta) ++ "]"
 ].
:- type state==set(item).
:- type encoded_state==int.



%% @see basecap#map_to_func2
/*:- func map_to_transitionTable(transitionTableMap(State)) = transitionTable(State)
	<= printable(State).
:- mode map_to_transitionTable(in) = out(transitionTable) is det.
map_to_transitionTable(TransitionTableMap) = ((func(State::in,Char::in) = (R::out) is semidet%%weakend
					      :-
					      R = map__search(TransitionTableMap, {State,Char})
					      ,log_print(">>\tlookup transition from " ++ to_string(State) ++ " under " ++ to_string(Char) ++ " to " ++ to_string(R) ++ " in\n\t" ++ to_string(TransitionTableMap) ++ "<<")
					      ; error("Lexical error: no transition in state " ++ to_string(State) ++ " under '" ++ to_string(Char) ++ "'")
					      )
					     ).
*/
/**
 * Generate acception map from states.
 * Every state containing an item [X->Alpha.] is a possible final state.
 * @todo detect reduce-reduce conflicts with friendly messages instead of just those by det_map_from_assoc_list
 */
:- func lr0_generateAcceptionMap(grammar(Token),set(state)) = acceptionMap(state,Token).
:- mode lr0_generateAcceptionMap(in,in) = out is semidet.
lr0_generateAcceptionMap(G,States) = ActionTableMap :-
	solutions(
		  (pred(pair(Q,Act)::out) is nondet :-
		      member(Q,States),
		      member(K,Q),
		      K = item(X, Alpha,[]),
		      member(lexical_rule(X, Alpha, Act), productions(G))
		  )
		 ,TableEntries
		  ),
	ActionTableMap = /*map__optimize*/(det_map_from_assoc_list_unprintable(TableEntries)).


/**
 * Create a scanner with LR(0) techniques. Produces a DFA not a DPDA
 * // G = {X1->r1,...,Xn->rn} is the regular grammar with occurrences of Xj in ri (j<i) instantiated
 * func start : state
 *     return {[X1->.r1],...,[Xn->.rn]}
 * 
 * // epsilon closure of s
 * // closure with respect to regular expression processing
 * func closure(s:state) : state
 * begin
 *     return fixedPoint(
 *         replace alpha.(r)*beta in s with alpha(.r)*beta, alpha(r)*.beta
 *         replace alpha(r.)*beta in s with alpha(.r)*beta, alpha(r)*.beta
 *         replace alpha.(r1|...|rn)beta in s with alpha(.r1|...|rn)beta, ..., alpha(r1|...|.rn)beta
 *         replace alpha(r1|...|ri.|...|rn)beta in s with alpha(r1|...|ri|...|rn).beta
 *     )
 * end
 * 
 * // shift transitions
 * func successors(s:state,a:Sigma):state
 * begin
 *     return {[X->alphaa.beta] with [X->alpha.abeta] in s}
 * end
 * 
 * var
 * q,q':state
 * Q:set(state)
 * d:state x Sigma -> state
 * begin
 *     Q := {closure(start)};
 *     d := {};
 *     foreachinc q in Q do
 *         for each a in Sigma do
 *             q' := closure(successors(q,a));
 * 	       if q' \= {} then
 * 	           if q' not in Q then
 * 	               Q := Q union {q'}
 * 	           fi
 * 	           d := d union {transition q->q' under a}
 * 	       fi
 *         od
 *     od
 * end
 */

:- func start(grammar(Token)) = state.
start(G) = list_to_set(map((func(lexical_rule(X,R,_Action)) = item(X, [],R)), productions(G))).

/**
 * // epsilon closure of s
 * // closure with respect to regular expression processing
 * func closure(s:state) : state
 * begin
 *     return fixedPoint(
 *         replace alpha.(r)*beta in s with alpha(.r)*beta, alpha(r)*.beta
 *         replace alpha(r.)*beta in s with alpha(.r)*beta, alpha(r)*.beta
 *         replace alpha.(r1|...|rn)beta in s with alpha(.r1|...|rn)beta, ..., alpha(r1|...|.rn)beta
 *         replace alpha(r1|...|ri.|...|rn)beta in s with alpha(r1|...|ri|...|rn).beta
 *     )
 * end
 */
:- func closure(grammar(Token),state) = state.
:- mode closure(in,in) = out is det.
closure(G,S) = R :-
	%% @internal this horrible implementation could by far be beautified and optimized
	R = fixedPoint(S,
		       (func(Q) = power_union(map(
						  (func(item(X,Alpha,Beta))
						  = ItemizedTransitions :-
						  ItemizedTransitions = map((func(Regexp) = Reitemized :-
							 Chars = string__to_char_list(Regexp),
							 nth_member_search(Chars, PositionMarker, Position),
							 split_list(Position-1, Chars, Head, [PositionMarker | Tail]),
							 Reitemized = item(X,Head,Tail)
							;
							 error("regular expression item without a position marker " ++ Regexp)
							),
							regexp_epsilon_transition(from_char_list(Alpha) ++ char_to_string(PositionMarker) ++ from_char_list(Beta))
						       )
						  ;
						      error("what to do when regexp parsing fails?")
						  ),
						  Q)
					     )
		       )
		      ),
	PositionMarker = (.).

/**
 * // shift transitions
 */
:- func successors(state,char) = state.
:- mode successors(in,in) = out is det.
successors(S,A) =
filter_map((func(item(X,Alpha,[A2|Beta])::in)
	   = (item(X,Alpha ++ [A2], Beta)::out) is semidet :-
	   A2 = A),
	   S
	  ).

/**
 * Generate 
 * @param G the grammar to generate a parser for.
 * @param States the current set of states.
 * @return NewStates the new set of states.
 * @param Transitions the current set of transitions pair({q,X},q') from q to q' under X.
 * @return NewTransitions the new set of transitions.
 * @param ToProcess the subset of States which still has to be processed via the foreach loop.
 * @param NewToProcess the remaining subset of states which still has to be processed.
 */
:- pred generate(grammar(Token),
		 set(state),set(pair({state,char},state)),
		 set(state),set(pair({state,char},state))).
:- mode generate(in,
		 in,in,
		 out,out) is det.
generate(G,
	 States0, Transitions0,
	 NewStates, NewTransitions) :-
	Sigma = character_classes(G),
	processEach((pred(S::in,ToProcess::out,
			  {States,Transitions}::in,{States2,Transitions2}::out) is semidet :-
		    News = filter_map(
				      (func(X::in) = (pair(Sp,pair({S,X},Sp))::out) is semidet :-
				          Sp = closure(G,successors(S,X)),
					  not set__empty(Sp)
				      %%,log_print("\nclosure(successors(" ++ to_string(X) ++ "," ++ to_string(S) ++ "))\n\t = closure(" ++ to_string(successors(S,X)) ++ ")\n\t = " ++ to_string(Sp) ++ "\n in states=" ++ to_string(States) ++ "\n")
				      )
				     , Sigma),
		     FoundStates = map(fst, News),
		     States2 = States `union` FoundStates,
		     Transitions2 = Transitions `union` map(snd, News),
		     %% only add those states to the process list that have been new to States.
		     ToProcess = to_sorted_list(filter(isnt(contains(States)), FoundStates))
		     ),
		    to_sorted_list(States0),
		    {States0,Transitions0}, {NewStates,NewTransitions}).


dfa_generate(G, TransitionTable, AcceptionMap, InitialState, CCG) :-
	CharacterGroups = union_list(map((func(lexical_rule(_,Regexp,_))
					  = R :-
					 R = regexp_char_groups(from_char_list(Regexp))
					 ; error("syntactically invalid regular expression " ++ from_char_list(Regexp))
					 ),
					 productions(G))),
	log_print("Character groups " ++ to_string(CharacterGroups)),
	(
	  log_print("Use " ++ to_string(char_group_filter(CharacterGroups, string_tokenSequence("ab0"))))
	),
	%% @todo disjunctify character groups
	%% @todo use resulting character groups in regular expressions treated like single characters
	
	Q0 = closure(G, start(G)),
	generate(G,
		 make_singleton_set(Q0), init,
		 States, Transitions),
	log_print("LR(0)Scan " ++ to_string(States)),
	InitialState1 = Q0,
	%% convert Transitions:set({{state,symbol},state}) to proper transitionTableMap
	TransitionTableMap1 = /*map__optimize*/(det_map_from_assoc_list(to_sorted_list(Transitions))),
	log_print("\ntransitionTable=\n" ++ to_string(TransitionTableMap1)),
	AcceptionMap1 = lr0_generateAcceptionMap(G, States),
	log_print("\nacceptionMap=\n" ++ to_string(keys(AcceptionMap1))),

	lr0_encode(States, InitialState1, TransitionTableMap1, AcceptionMap1,
		   InitialState, TransitionTableMap, AcceptionMap),

	log_print("\ntransitionTable=\n" ++ to_string(TransitionTableMap)),
	log_print("\nacceptionMap=\n" ++ to_string(keys(AcceptionMap))),

	CCG = id,
	TransitionTable = TransitionTableMap.
	%%TransitionTable = map_to_transitionTable(TransitionTableMap).


/**
 * Encodes the states of an LR(0) generated scanner.
 */
:- all [State] /*some [EncodedState]*/ ((
	pred lr0_encode(set(State), State, transitionTableMap(State,CharClass), acceptionMap(State,Token),
						   encoded_state, transitionTableMap(encoded_state,CharClass), acceptionMap(encoded_state,Token))
				    %%=> printable(EncodedState)
				    ) <= (printable(State),token(Token),printable(CharClass))).
:- mode lr0_encode(in, in,in,in, out,out,out) is det.

lr0_encode(States,
	   InitialState, TransitionTable, AcceptionMap,
	   EncInitialState, EncTransitionTable, EncAcceptionMap) :-
	StatesList = to_sorted_list(States),
	Encode = (func(S) = I-1
		 :-
		     nth_member_search(StatesList,S,I)
		 ;
		     error("State not found")
		 ) `with_type` (func(State)=int),
	EncInitialState = Encode(InitialState),
	EncTransitionTable = /*map__optimize*/(det_map_from_assoc_list(map(
						  semidet_to_det((func(({S,A}-Sp)::in) = (({Encode(S),A}-Encode(Sp))::out) is semidet),
								  "transitionTableMap type"),
						  to_assoc_list(TransitionTable)
						 ))),
	EncAcceptionMap = /*map__optimize*/(det_map_from_assoc_list_unprintable(map(
						semidet_to_det_unprintable((func((S-Act)::in) = ((Encode(S)-Act)::out) is semidet),
							       "acceptionMap type"),
						to_assoc_list(AcceptionMap)
					       ))).

dfa_scanner(G) = S :-
	dfa_generate(G, Tau, F, S0, CCG),
	S = dfa_scanner(Tau, F, S0, CCG).

:- end_module scanner_generator__lr0Generator.