/**
 * Provides scanner generators.
 * @author Andre Platzer
 * @version 2003-08-20
 */
:- module scanner_generator.
:- interface.
:- include_module lr0Generator.

:- import_module scanners, basecap,
	char, list, set.

/**
 * An atomic regular expression "literal".
 * @author Andre Platzer
 * @version 2003-08-20
 */
:- type atomic_regular_expression == char.

/**
 * A CH-3 regular expression.
 * @author Andre Platzer
 * @version 2003-08-20
 */
:- type regular_expression == list(atomic_regular_expression).

/**
 * Lexical actions like parser functions on actionTypes.
 * <= token(Token)
 */
:- type lexical_action(Token) == (func(string) = Token).

/**
 * A lexical CH-3 production.
 * lexicalrule(X, alpha, action) represents the regular CH-3 production
 * <center>X&rarr;alpha</center>
 * with the given lexical action to happen on each application of the production.
 * When having applied the production and when alpha matched the string Str
 * the lexer invokes X.token = action(Str).
 * @author Andre Platzer
 * @version 2003-08-20
 */
:- type production(Token)
	---> lexical_rule(
			  token_kind,
			  regular_expression,
			  lexical_action(Token)
			 ).

:- instance printable(production(Token)) <= printable(Token).

/**
 * Create a lexical production with default identical token creation.
 */
:- (func new_rule(token_kind, regular_expression, lexical_action(Token)) = production(Token)
	<= token(Token)).

/**
 * Represents a regular CH-3 grammar.
 * @param Token the type of tokens produces by the lexical actions.
 * @author Andre Platzer
 * @version 2003-08-20
 */
:- type grammar(Token)
	---> grammar(
		     productions :: list(production(Token))
		    ).

/**
 * Determines the character classes used in a grammar G.
 */
:- func character_classes(grammar(Token)) = set(char).
:- mode character_classes(in) = out is det.



:- implementation.
:- import_module string.

:- instance printable(production(Token)) <= printable(Token) where [
	to_string(lexical_rule(TK, R, Action)) = to_string(TK) ++ " --> " ++ to_string(R)
 ].


character_classes(G) = list_to_set(condense(map(
						(func(lexical_rule(X,Alpha,_)) = Alpha),
						productions(G)
					       ))).

new_rule(TK, R, Action) = lexical_rule(TK, R, Action).

:- end_module scanner_generator.