/**
 * Provides diverse scanner utilities.
 * @author Andre Platzer
 * @version 2003-08-20
 */
:- module scanner_util.
:- interface.
:- import_module scanners, scanner_generator, basecap, list.


:- instance tokenSequence(list(Token)) <= token(Token).


/**
 * Create a lexical production with default identical token creation.
 */
:- some [Token] (func new_rule(token_kind, regular_expression) = production(Token)
	=> token(Token)).

/**
 * Canonical token representations.
 * @author Andre Platzer
 * @version 2003-08-20
 */
:- type canonical_token.
:- instance printable(canonical_token).
:- instance token(canonical_token).

/**
 * Create a lexical production with default identical token creation.
 * Different images will have different kinds.
 */
:- func rule(token_kind, regular_expression) = production(canonical_token).

/**
 * Create a lexical production with default identical token creation
 * abstracting all tokens of different images into one group of the same kind.
 * Different images will have same kinds.
 */
:- func rule_group(token_kind, regular_expression) = production(canonical_token).

/**
 * Create a lexical skip rule, skipping all occurrences of the given regular expressions.
 */
:- func skip(regular_expression) = production(canonical_token).



:- implementation.
:- import_module string, require.

:- instance tokenSequence(list(Token)) <= token(Token) where [
	(get_token(Tok, In, Out) :-
	     (
	       In = [],
	       Tok = end_of_file,
	       Out = In
	     ;
	       [Tok | Out] = In
	     )
	),

	%%unget_token(T, Ts, [T | Ts]),
	empty([end_of_file]),
	empty([]) :- error("[] not expected within token sequence. Should stick to eof before")
].

:- type canonical_token
	---> canonical_token(token_kind)
	;    canonical_token_class(token_kind, string).
:- instance printable(canonical_token) where [
	to_string(canonical_token(S)) = to_string(S),
	to_string(canonical_token_class(T,S)) = to_string(T) ++ "[" ++ S ++ "]"
 ].

:- instance token(canonical_token) where [
        get_value(canonical_token_class(X,V)) = V,
        get_value(canonical_token(X)) = _ :- error("cannot get value of a token_kind anyway"),
        get_kind(canonical_token(X)) = X,
        get_kind(canonical_token_class(X,V)) = X,
	end_of_file = canonical_token(end_of_file)
 ].

new_rule(TK, R) = lexical_rule(TK, R,
			       (func(S) = make_token_kind(S))).

rule(TK, R) = lexical_rule(TK, R,
			   (func(S) = canonical_token(make_token_kind(S)))).

rule_group(TK, R) = lexical_rule(TK, R,
				      (func(S) = canonical_token_class(TK, S))).

skip(R) = rule(make_token_kind("SKIP"), R).

:- end_module scanner_util.
