/**
 * Provides a scanner based on a deterministic finite automaton.
 * @author Andre Platzer
 * @version 2003-08-20
 */
:- module scanners__dfa.
:- interface.
:- import_module scanner_generator, map.

/**
 * A state transition table for finite automata.
 * The transition table associates with each state q and transition character c
 * of class C the state to which the acceptorshould make a transition.
 * @preconditions q:State & c:char & C:CharClass
 *  & mode actionTable(q::in,C::in)=out is semidet
 * @xxx we do not currently use functions, because their higher-order insts differences cause
 *  serious runtime bugs crashes and silent data reinterpretation causing loss of type safety.
 */
%%:- type transitionTable(State,CharClass) == (func(State,CharClass) = State).
%%:- inst transitionTable == (func(in,in) = out is semidet).
:- type transitionTable(State,CharClass) == (map({State,CharClass},State)).

/**
 * A state acception map for finite automata.
 * A pair (q-a) of an acception map specifies that q is a final state and that
 * a(string_image_of_token) is the resulting output token.
 */
:- type acceptionMap(State,Token) == map(State,lexical_action(Token)).

/**
 * Groups characters into their character classes.
 * For example may group b to its character class [a-z] of letters.
 * @xxx will no longer be necessary, once mercury fixes its higher-order inst mode
 *  problem such that transitionTables can be functions again.
 */
:- type charClassMapper(CharClass) == (func(char)=CharClass).


/**
 * Create a scanner based on a deterministic finite automaton.
 * @param transitionTable the transition table of deterministic finite automaton
 *  determining the transition under some symbol in a particular state.
 * @param acceptionMap the acception map of the deterministic finite acceptor
 *  determining which states are accepting states, and what the outputs tokens are.
 * @param initialState the initial state of the DFA to start parsing.
 * @param charClassMapper will be used to map individual characters to their
 *  character class before making transitions.
 */
:- all [State,Token,CharClass] some [S] (func dfa_scanner(
						transitionTable(State,CharClass),
						acceptionMap(State,Token),
						State,
						charClassMapper(CharClass)) = S
			=> scanner(S))
        <= (token(Token),printable(State),printable(CharClass)).
:- mode dfa_scanner(in/*(transitionTable)*/, in, in, in) = out is det.



:- implementation.

:- import_module list, string, char, require, std_util, basecap.
:- import_module string_source.


%% dfa implementation

:- type deterministic_finite_automaton(State,Token,CharClass)
	---> dfa(
		 transitions :: transitionTable(State,CharClass),
		 acceptions :: acceptionMap(State,Token),
		 initialState :: State,
		 characterClassGrouper :: charClassMapper(CharClass)
		).
%%:- inst deterministic_finite_automaton == bound(dfa(transitionTable,ground,ground)).

:- type incremental_tokenSequence(TSIn,State,Token,CharClass)
	---> incremental_tokenSequence(
				       deterministic_finite_automaton(State,Token,CharClass),
				       TSIn
				      ).
%%:- inst incremental_tokenSequence ==
%%	bound(incremental_tokenSequence(deterministic_finite_automaton,ground)).
:- instance tokenSequence(incremental_tokenSequence(TSIn,State,Token,CharClass))
	<= (printable(State),token(Token),tokenSequence(TSIn)).

:- instance printable(deterministic_finite_automaton(State,Token,CharClass))
	<= (printable(State),token(Token)) where [
	to_string(dfa(Tau,F,S0,CCG)) = "DFA(initialState=" ++ to_string(S0) ++ ", acception=" ++ to_string(keys(F)) ++ ")"
  ].

:- instance scanner(deterministic_finite_automaton(State,Token,CharClass))
	<= (printable(State),token(Token)) where [
	scan(S, Input) = incremental_tokenSequence(S, Input)
  ].

dfa_scanner(Tau, F, S0, CCG) = dfa(Tau, F, S0, CCG).


/**
 *
 * state = S0
 * C = read_char()
 * while (state not in F) do
 *     state := transition(state, C)
 *     C := read_char()
 * od
 * if (state = error)
 *     error
 * else
 *     return text accumulated from C this run
 * fi
 */
:- pred myget_token(Token,
		    incremental_tokenSequence(TSIn,State,Token,CharClass),
		    incremental_tokenSequence(TSIn,State,Token,CharClass))
        <= (token(Token),printable(State),tokenSequence(TSIn)).
:- mode myget_token(out,
		    in,%%in(incremental_tokenSequence),
		    out%%out(incremental_tokenSequence)
		    ) is det.
myget_token(Tok,
	    incremental_tokenSequence(DFA, In),
	    incremental_tokenSequence(DFA, Out)) :-
	(
	  accept(DFA,
		 initialState(DFA), [],
		 Tok,
		 In, Out)
	;
	  error("lexical error somewhere")
	).

:- pred accept(deterministic_finite_automaton(State,Token,CharClass),
	       State, list(char),
	       Token,
	       TSIn, TSIn)
        <= (token(Token),printable(State), tokenSequence(TSIn)).
:- mode accept(in,
	       in, in,
	       out,
	       in, out) is semidet.
accept(DFA,
       State, Image,
       Tok,
       In, Out) :-
	%%%log_print("\t   search " ++ to_string(State) ++ " in finals " ++ to_string(keys(F))),
	( Action = search(acceptions(DFA), State) ->
	    %% accepted state
	    TokRecognized = Action(string__from_rev_char_list(Image)),
	    %%%log_print("\taccepted " ++ to_string(State) ++ " emitting " ++ to_string(Tok)),
	    (
	      %% try to continue to longest possible match
	      (
		get_token(Char, In, Cont),
		char_token(C) = cast_det(Char,"character-base stream expected"),
		search(transitions(DFA),{State,characterClassGrouper(DFA)(C)}, Statep),
		accept(DFA,
		       Statep, [C | Image],
		       TokLonger,
		       Cont, OutLonger)
	      ) ->
	      Tok = TokLonger,
	      Out = OutLonger
	    ;
	      %% already is longest possible match
	      (get_kind(TokRecognized) = skip ->
		  %% @internal built-in skip handling (instead of relying on the siever)
		  %% restart tokenization from current position
		  accept(DFA,
			 initialState(DFA), [],
			 Tok,
			 In, Out)
	      ;
		  Tok = TokRecognized,
		  Out = In
	      )
	    )
	;
	    %% state that has to be continued for reaching a valid token
	    ( (get_token(Char, In, Cont),
	       char_token(C) = cast_det(Char,"character-base stream expected")
	      ) ->
		%%%log_print("\tshift " ++ to_string(C)),
		%%%%Statep = Tau(State,C),
		(
		  search(transitions(DFA),{State,characterClassGrouper(DFA)(C)}, Statep)
		;
		  error("Lexical error: no transition in state " ++ to_string(State) ++ " under '" ++ to_string(C) ++ "'")
		),
		%%%log_print("\ttransition from " ++ to_string(State) ++ " to " ++ to_string(Statep) ++ " under " ++ to_string(C)),
		accept(DFA,
		       Statep, [C | Image],
		       Tok,
		       Cont, Out)
	    ;
		%% read failed
		(Image = [] ->
		    Tok = end_of_file,
		    Out = In
		;
		    error("Unexpected characters at end of file: " ++ to_string(Image))
		)
	    )
	).

:- instance printable(incremental_tokenSequence(TSIn,State,Token,CharClass))
	<= (printable(State),token(Token),tokenSequence(TSIn)) where [
	to_string(TS)
	 = "incremental_tokenSequence(" ++ get_all_tokens_string(TS) ++ ")"
].

:- instance tokenSequence(incremental_tokenSequence(TSIn,State,Token,CharClass))
	<= (printable(State),token(Token),tokenSequence(TSIn)) where [
	pred(get_token/3) is myget_token,

	%%unget_token(T, Ts, [T | Ts]),
	(empty(T) :- get_token(end_of_file,T,_))
].

:- end_module scanners__dfa.