/**
 * Provides abstract scanner data representations.
 * @author Andre Platzer
 * @version 2003-08-01
 */
:- module scanners.
:- interface.
:- include_module dfa.

:- import_module list, string, basecap, std_util.

/**
 * The kind or class of lexical token representations, independent from
 * the particular value.
 * For parsing, only the kind of a token is important, not (directly) the value. 
 * @author Andre Platzer
 * @version 2003-08-17
 */
:- type token_kind.

/**
 * The type class for lexical token representations.
 * @author Andre Platzer
 * @version 2003-08-16
 */
:- typeclass token(Token) <= printable(Token) where [
        /**
	 * Get the content value of a terminal.
	 */
	some [ValueType] (func get_value(Token) = ValueType
			 => printable(ValueType)),
	mode get_value(in) = out is det,

        /**
	 * Get the kind or class of a token.
	 */
	/*some [TokenKind]*/ func get_kind(Token) = token_kind,
	mode get_kind(in) = out is det,

       /**
	 * The end of file token.
	 */
	func end_of_file = Token,
	mode end_of_file = out is det
  ].

/**
 * Check tokens for equal token class.
 * @see token#get_kind/1
 */
:- all [Token1,Token2] (pred token_class_equals(Token1,Token2)
		       <= (token(Token1),token(Token2))).
:- mode token_class_equals(in,in) is semidet.

:- type word(Token) == list(Token).


/**
 * token_kinds behave like tokens, especially they support end_of_file
 */
:- instance token(token_kind).
:- instance printable(token_kind).
:- type word_kind == list(token_kind).


/**
 * A token sequence or token stream obtained from a scanner.
 * @author Andre Platzer
 * @version 2003-08-01
 * @xxx perhaps we should have added an explicit token type paramter Token instead of some [Token].
 *  However, this attempt has caused manious compilation problems.
 */
:- typeclass tokenSequence(TokenSequence) <= printable(TokenSequence) where [
	/**
	 * empty(DCGin)
	 * Check whether token stream is empty.
	 */
	pred empty(TokenSequence),
	mode empty(in) is semidet,
	/**
	 * get_token(Tok, DCGin, DCGout)
	 * Return the next token Tok in the token stream
	 * from the lexer state DCGin,
	 * ending in the lexer state DCGout.
	 */
	some [Token] (pred get_token(Token, TokenSequence, TokenSequence)
		=> token(Token)),
	mode get_token(out, in, out) is det
	/**
	 * unget_token(Tok, DCGin, DCGout)
	 * Put the token Tok back to the token stream
	 * from the lexer state DCGin,
	 * ending in the lexer state DCGout.
	 */
	/*some [Token] (pred unget_token(Token, TokenSequence, TokenSequence)
		=> token(Token)),
	mode unget_token(in, in, out) is det
	%% invariant? all [Tok, T0] some [T] get_token(Tok, unget_token(Tok, T0), T)
	  */
].


/**
 * A scanner typeclass.
 * @param Scanner the concrete scanner type.
 * @author Andre Platzer
 * @version 2003-08-04
 * @todo distinguish TS containing only terminals, from stack, containing terminals or other symbols.
 */
:- typeclass scanner(Scanner) <= printable(Scanner) where [
	/**
	  * scan(Scanner, Input) = TS
	  * Let Scanner scan the input stream producing as output a
	  * token sequence.
	  * @todo generalize input to general input streams.
	  * @todo transform into something that can either read all, or read one symbol
	  */
	all [TSIn] some [TS] (func scan(Scanner, TSIn) = TS => tokenSequence(TS))
		<= tokenSequence(TSIn),
	mode scan(in, in) = out is semidet
].


/**
 * Create a token_kind of the given class description.
 * @todo could generalize argument to all [T] T <= printable(T)
 */
:- func make_token_kind(string) = token_kind.


%% peculiarities

/**
 * The token_kind for tokens that should be left out during siever and never
 * delivered to the parser. Whitespaces typically fall into this kind.
 */
:- func skip = token_kind.


%% filtering and sievers

:- func foldr((func(Token, Y) = Y), Y, TS) = Y
	<= (tokenSequence(TS), token(Token)).
:- mode foldr(in(func(in, in) = out is det), in, in) = out is det.
%%:- mode foldr(in(func(in, di) = uo is det), di, uo) = out is det.
:- mode foldr(in(func(in, in) = out is semidet), in, in) = out is semidet.
/*:- mode foldr(in(func(in, in) = out is nondet), in, in) = out is nondet.
:- mode foldr(in(func(in, di) = uo is cc_multi), in, di) = uo is cc_multi.
:- mode foldr(in(func(in, in) = out is cc_multi), in, in) = out is cc_multi.
*/

:- func foldl((func(Token, Y) = Y), Y, TS) = Y
	<= (tokenSequence(TS), token(Token)).
:- mode foldl(in(func(in, in) = out is det), in, in) = out is det.
%%:- mode foldl(in(func(in, di) = uo is det), di, uo) = out is det.
:- mode foldl(in(func(in, in) = out is semidet), in, in) = out is semidet.
/*:- mode foldl(in(func(in, in) = out is nondet), in, in) = out is nondet.
:- mode foldl(in(func(in, di) = uo is cc_multi), in, di) = uo is cc_multi.
:- mode foldl(in(func(in, in) = out is cc_multi), in, in) = out is cc_multi.
*/


/**
 * filter_map(Transformer, TokS, TrueTokS) takes a predicate
 * with one input argument and one output argument. It is called
 * with each element of TokS. If a call succeeds, then the output is
 * included in TrueTokS.
 * Can be used for implementing sievers.
 * @xxx semideterminism of Transformer will most probably cause mercury crash at runtime.
 * @xxx univ for argument type of Transformer seems necessary. Otherwise mercury crashes at runtime.
 */
:- all [TokenSequence] some [FilteredTokenSequence] (
		func filter_map(func(Token)=FilteredToken, TokenSequence) = FilteredTokenSequence
		=> tokenSequence(FilteredTokenSequence)
	)
	<= (token(Token), token(FilteredToken), tokenSequence(TokenSequence)).
:- mode filter_map(func(in) = out is semidet, in) = out is det.

:- all [TokenSequence,FilteredToken] some [FilteredTokenSequence] (
		func map_univ(func(Token)=FilteredToken, TokenSequence) = FilteredTokenSequence
		=> tokenSequence(FilteredTokenSequence)
	)
	<= (tokenSequence(TokenSequence), token(Token), token(FilteredToken)).
:- mode map_univ(func(in) = out is det, in) = out is det.




:- implementation.

:- import_module list, string, char, require, std_util.
	
:- type token_kind
	---> token_kind(
		  name :: string
		 ).

make_token_kind(X) = token_kind(X).

:- instance printable(token_kind) where [
	to_string(token_kind(X)) = "<" ++ X ++ ">"
 ].

:- instance token(token_kind) where [
        get_value(_) = "" :- error("cannot get value of a token_kind"),
        func(get_kind/1) is id,
	end_of_file = make_token_kind("end_of_file")
 ].

token_class_equals(A,B) :-
	get_kind(A) = get_kind(B).


:- func get_all_tokens_string(TS) = string <= tokenSequence(TS).
get_all_tokens_string(In) = R :-
	get_all_tokens_stringHelper(R,In,_).

:- pred (get_all_tokens_stringHelper(string::out,TS::in,TS::out) is det)
	<= tokenSequence(TS).
get_all_tokens_stringHelper(R) -->
	get_token(Tok),
	%%%{log_print("\tfound " ++ to_string(Tok))},
	( { Tok = end_of_file } ->
	    { R = "" }
	; 
	    get_all_tokens_stringHelper(R2),
	    {R = to_string(Tok) ++ R2}
	).

foldr(F, C, Toks) = R :-
	get_token(Tok, Toks, ToksRest),
	( Tok = end_of_file ->
	    R = C
	;
	    R = F(cast_det(Tok, "argument type cast failed"), foldr(F, C, ToksRest))
	).

foldl(F, Toks, C) = R :-
	error("not yet implemented").

%% filtering

:- type filteredTokenSequence(Token,TokenSequence,FilteredToken)
	---> filtered(func(Token)=FilteredToken, TokenSequence).
:- instance printable(filteredTokenSequence(Token,TokenSequence,FilteredToken))
        <= printable(TokenSequence) where [
      to_string(filtered(_, TokS)) = "filtered " ++ to_string(TokS)
 ].
:- instance tokenSequence(filteredTokenSequence(Token,TokenSequence,FilteredToken))
        <= (token(Token), token(FilteredToken), tokenSequence(TokenSequence)) where [
	(get_token(Tok, filtered(F,In), filtered(F,Out)) :-
             get_token(ATok, In, Mid),
	     (
	       %% @xxx warning determinism information about F lost. So expected to be det?
	       (Result = F(cast_det(ATok, "could not cast token to type expected by filter") `with_type` Token) ->
		   Tok = cast_det(Result, "could not cast result of filter to expected type"),
		   Out = Mid
	       ;
		   get_token(RTok, Mid, Out),
		   Tok = cast_det(RTok, "this cast should never fail due to recursive case")
	       )
	     )
	),
	(empty(In) :- get_token(end_of_file, In, _))
].
  
filter_map(F, TokS) = filtered(F,TokS) :-
	log_print("warning, scanners__filter_map does not yet work due to mercury bug").


:- type map_univTokenSequence(TokenSequence,Token,MapToken)
	---> mapped_univ(func(Token)=MapToken, TokenSequence).
:- instance printable(map_univTokenSequence(TokenSequence,Token,MapToken))
        <= (tokenSequence(TokenSequence), token(Token), token(MapToken)) where [
      to_string(mapped_univ(F, TokS)) = "map_univ " ++ to_string(TokS) ++ "\n\tis" ++ get_all_tokens_string(mapped_univ(F, TokS))
 ].
:- instance tokenSequence(map_univTokenSequence(TokenSequence,Token,MapToken))
        <= (tokenSequence(TokenSequence), token(Token), token(MapToken)) where [
	(get_token(Tok, mapped_univ(F,In), mapped_univ(F,Out)) :-
             get_token(ATok, In, Mid),
	     (
	       Result = F(cast_det(ATok, "could not cast argument of filter to expected type")),
	       Tok = Result,
	       Out = Mid
	     )
	),
	(empty(In) :- get_token(end_of_file, In, _))
].
  
map_univ(F, TokS) = mapped_univ(F,TokS).


:- pragma memo(skip/0).
skip = make_token_kind("SKIP").

:- end_module scanners.