/**
 * Provides a simple tokenizer.
 * @author Andre Platzer
 * @version 2003-08-01
 */
:- module simple_tokenizer.
:- interface.
:- import_module scanners.

/**
 * Create a simple tokenizer separating alphanumeric identifiers from numbers and operators.
 * @attribute implementation simplistic
 */
:- some [S] (func simple_tokenizer = S => scanner(S)).
:- mode simple_tokenizer = out is det.



:- implementation.

:- import_module list, string, char, require, std_util, basecap.
:- import_module token_impl, scanner_util, string_source.

:- all [TSIn] some[TS] (func scan(TSIn) = TS => tokenSequence(TS))
	<= tokenSequence(TSIn).
:- mode scan(in) = out is semidet.


:- type simpleTokenizer
	---> simpleTokenizer.

:- instance printable(simpleTokenizer) where [
	to_string(_) = "simpleTokenizer tokenizes simply alphanumerics, numbers, operators and delimiters"
  ].
:- instance scanner(simpleTokenizer) where [
	scan(S, Input) = scan(Input)
  ].

simple_tokenizer = simpleTokenizer.


/**
 * The types of lexical symbol or token representations.
 * @author Andre Platzer
 * @version 2003-08-01
 */
:- type token.
:- instance printable(token).
:- instance token(token).


:- type token == token_impl.

scan(Input) = list__reverse(Tokens) :-
	scan(foldr((func(Tok `with_type` char_token,Chars) = [C|Chars] :-
				 char_token(C) = cast_det(Tok, "character-based stream expected")
				 ;
				  error("this should never happen")
		   ), [], Input),
	     [], Tokens).
:- pred scan(list(char), list(token), list(token)).
:- mode scan(in, in, out) is det.
scan([], Toks, [endoffile | Toks]).
scan([C | Cs], Toks0, Toks) :-
    (
        char__is_whitespace(C)
    ->
        scan(Cs, Toks0, Toks)
    ;
        char__is_alpha(C)
    ->
        takewhile(char__is_alpha, [C | Cs], Characters, Rest),
        string__from_char_list(Characters, String),
        scan(Rest, [identifier(String) | Toks0], Toks)
    ;
        char__is_digit(C)
    ->
        takewhile(char__is_digit, [C | Cs], Digits, Rest),
        string__from_char_list(Digits, NumStr),
        Num = string__det_to_int(NumStr),
        scan(Rest, [identifier(int_to_string(Num)) | Toks0], Toks)
    ;
        C = '\"'
    ->
        takewhile((pred(Char::in) is semidet :-
		  (Char = '\r' ; Char = '\n') ->
		   error("lexical error, string literal extends over end of line")
		  ;
		   Char \= '\"'
		  ), Cs, Characters, Rest),
        (
	  Rest = ['\"' | Rest2]
	;
	  error("lexical error, unterminated string literal")
	),
        scan(Rest2, [string_literal(string__from_char_list(Characters)) | Toks0], Toks)
    ;
        (C = (+) ; C = (*))
    ->
        scan(Cs, [operator(char_to_string(C)) | Toks0], Toks)
    ;
        (C = (-) ; C = (/) ; C = (<) ; C = (>))
    ->
        scan(Cs, [operator(char_to_string(C)) | Toks0], Toks)
    ;
        (C = (.) )
    ->
        scan(Cs, [operator(char_to_string(C)) | Toks0], Toks)
    ;
        (C = '[' ; C = ']' ; C = '(' ; C = ')' ; C = (',') ; C = (;) ; C = (:))
    ->
        scan(Cs, [delimiter(char_to_string(C)) | Toks0], Toks)
    ;
        (C = (!) ; C = (@) ; C = '#' ; C = '$' ; C = '%' ; C = ('&') ; C = ('^') ; C = '|')
    ->
        scan(Cs, [delimiter(char_to_string(C)) | Toks0], Toks)
    ;
        error("expr: lexical syntax error in input at " ++ char_to_string(C))
    ).

:- end_module simple_tokenizer.